# -*- coding:  utf-8 -*-
"""
Created on 2020/8/5
@project: SPAIC
@filename: Connection
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义神经集群间的连接，包括记录神经元集群、连接的突触前、突触后神经元编号、连接形式（全连接、稀疏连接、卷积）、权值、延迟 以及连接产生函数、重连接函数等。
"""
from spaic.Network.Topology import Connection
from spaic.Network.Assembly import Assembly
import numpy as np
import scipy.sparse as sp
import torch
from spaic.IO.Initializer import BaseInitializer
import collections
import random
# from spaic.utils.memory import get_cpu_mem,get_object_size

from matplotlib.pyplot import *
import math

def _pair(x):
    if x is None:
        return None
    elif isinstance(x, collections.abc.Iterable):
        return tuple(x)
    else:
        assert isinstance(x, int)
        return (x, x)
class FullConnection(Connection):

    '''
    each neuron in the first layer is connected to each neuron in the second layer.

    Args:
        pre(Assembly): The assembly which needs to be connected
        post(Assembly): The assembly which needs to connect the pre
        link_type(str): full
    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(FullConnection, self).__init__(pre=pre, post=post, name=name,
                                             link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                             sparse_with_mask=sparse_with_mask,
                                             pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

        if weight is None:
            # Connection weight
            if self.post.model_name == 'double_complex':
                self.shape = list(self.shape) + [12] # 3*2 dimensional attention
                self.weight = self.w_std*np.random.randn(*self.shape) + self.w_mean
            else:
                self.weight = self.w_std*np.random.randn(*self.shape) + self.w_mean

        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight


        self.Isyn_bn = kwargs.get('Isyn_bn', False)
        if self.Isyn_bn is True:
            self.Is = []
            self.beta_ave = kwargs.get('beta_ave', 0.99)
            self.beta_std = kwargs.get('beta_std', 0.999)
            self.beta_bn = kwargs.get('beta_bn', 0.001)
            self.targ_ave = kwargs.get('targ_ave', 1.5)
            self.targ_std = kwargs.get('targ_std', 2.0)
            self.running_ave = self.targ_ave
            self.running_std = self.targ_std
            self.running_bias = 0.0
            self.running_scale = 1.0

            self._operations.append([None, self.update_ave, 'V'+'[post]'])
            self._init_operations.append(['weight[link]', self.update_bn, [post_var_name+'[post]', 'weight[link]']])

        # construct unit connection information by policies,
        # construct in __init__ is potentially bad, as network structure may change before build. should add new function
        # self.connection_inforamtion = ConnectInformation(self.pre, self.post)
        # self.connection_inforamtion.expand_connection()
        # for p in self._policies:
        #     self.connection_inforamtion = self.connection_inforamtion & p.generate_connection(self.pre, self.post)

    def update_ave(self, Isyn):
        with torch.no_grad():
            if Isyn.dtype.is_complex:
                self.Is.append(Isyn.imag.detach())
            else:
                self.Is.append(Isyn.detach())

    def update_bn(self, Isyn, weight):
        if self.training:
            weight.register_hook(self.norm_hook)
            with torch.no_grad():
                if len(self.Is) > 1:
                    Is = torch.cat(self.Is, dim=0)
                    mean_rate = torch.mean(Is, dim=0)
                    std = torch.std(Is, dim=0)
                    self.running_ave = self.beta_ave*self.running_ave + (1-self.beta_ave)*mean_rate
                    self.running_std = self.beta_std*self.running_std + (1-self.beta_std)*std

                    d_bias = (self.targ_ave - self.running_ave)*torch.sigmoid(20.0*(self.targ_ave - self.running_ave)*(self.running_ave-mean_rate))
                    self.running_bias = self.running_bias + self.beta_bn*d_bias
                    d_scale = torch.exp((self.targ_std - self.running_std)) - 1.0
                    self.running_scale = self.running_scale + self.beta_bn*d_scale
                    self.Is = []
                    self.weight_change = True
                else:
                    self.weight_change = False
                    self.running_scale = torch.ones_like(torch.mean(Isyn.real, dim=0))
                    self.running_bias = torch.zeros_like(torch.mean(Isyn.real, dim=0))
        else:
            self.Is = []

        if self.weight_change:
            weight = self.running_scale.unsqueeze(-1) * weight + self.running_bias.unsqueeze(-1)
            if weight.requires_grad:
                weight.retain_grad()


        return weight
    def norm_hook(self, grad):
        mean_grad = torch.mean(grad, dim=1, keepdim=True)
        return grad - mean_grad

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass

Connection.register('full', FullConnection)
Connection.register('full_connection', FullConnection)

class one_to_one_sparse(Connection):
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(one_to_one_sparse, self).__init__(pre=pre, post=post, name=name,
                                                link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                sparse_with_mask=sparse_with_mask,
                                                pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        try:
            assert self.pre_num == self.post_num
        except AssertionError:
            raise ValueError(
                'One to One connection must be defined in two groups with the same size, but the pre_num %s is not equal to the post_num %s.' % (
                self.pre_num, self.post_num))
        weight = kwargs.get('weight', None)
        self.w_mean = kwargs.get('w_mean', 0.05)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', False)
        self.is_sparse = kwargs.get('is_sparse', True)

        if weight is None:
            # Connection weight
            self.weight = self.w_mean * np.eye(*self.shape)
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError('Sparse implementation of one_to_one_sparse connection do not support the %s '
                                 'initialization method. You can pass a numpy matrix whose diagonal is not zero.',
                                 self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight
        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value


    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass

Connection.register('one_to_one_sparse', one_to_one_sparse)


class one_to_one_mask(Connection):
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=True, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(one_to_one_mask, self).__init__(pre=pre, post=post, name=name,
                                              link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                              sparse_with_mask=sparse_with_mask, pre_var_name=pre_var_name,
                                              post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        try:
            assert self.pre_num == self.post_num
        except AssertionError:
            raise ValueError('One to One connection must be defined in two groups with the same size, but '
                             'the pre_num %s is not equal to the post_num %s.' % (self.pre_num, self.post_num))
        weight = kwargs.get('weight', None)
        self.w_mean = kwargs.get('w_mean', 0.05)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        # self._variables = dict()

        if weight is None:
            # Connection weight
            self.weight = self.w_mean * np.eye(*self.shape)
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError('Dense implementation of one_to_one_mask connection do not support the %s '
                                 'initialization method. You can pass a numpy matrix whose diagonal is not zero.', self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('one_to_one', one_to_one_mask)

class conv_connect(Connection):

    '''
    do the convolution connection.

    Args:
        pre(Assembly): the assembly which needs to be connected
        post(Assembly): the assembly which needs to connect the pre
        link_type(str): Conv
    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['conv'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['conv']
        super(conv_connect, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        self.out_channels = kwargs.get('out_channels', None)
        self.in_channels = kwargs.get('in_channels', None)
        self.kernel_size = kwargs.get('kernel_size', [3, 3])
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.05)
        weight = kwargs.get('weight', None)

        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        self.mask = kwargs.get('mask', None)
        self.stride = kwargs.get('stride', 1)
        self.padding = kwargs.get('padding', 0)
        self.dilation = kwargs.get('dilation', 1)
        self.groups = kwargs.get('groups', 1)
        self.upscale = kwargs.get('upscale', None)

        if isinstance(self.kernel_size, int):
            self.kernel_size = [self.kernel_size] * 2
        if isinstance(self.stride, int):
            self.stride = [self.stride] * 2
        if isinstance(self.padding, int):
            self.padding = [self.padding] * 2
        # self._variables = dict()
        self._constant_variables['stride[link]'] = self.stride
        self._constant_variables['padding[link]'] = self.padding
        self._constant_variables['dilation[link]'] = self.dilation
        self._constant_variables['groups[link]'] = self.groups
        if self.upscale is not None:
            self._constant_variables['upscale[link]'] = self.upscale

        if self.in_channels is None:
            from spaic.Neuron.Encoders import NullEncoder
            if len(self.pre.shape) == 1:
                raise ValueError('The shape %s of pre seems to be a flattened value. '
                                 'Give the correct shape with structure [channel, height, width] in the NeuronGroup '
                                 'class with the neuron_shape parameter' % self.pre.shape)
            elif isinstance(self.pre, NullEncoder):
                self.in_channels = self.pre.shape[2]
            else:
                self.in_channels = self.pre.shape[0]

        if self.out_channels is None:
            if len(self.post.shape) == 1:
                raise ValueError('The shape %s of post seems to be a flattened value.  '
                                 'Give the correct shape with structure [channel, height, width] in the NeuronGroup '
                                 'class with the neuron_shape parameter' % self.post.shape)
            else:
                self.out_channels = self.post.shape[0]

        self.shape = (self.out_channels, self.in_channels, self.kernel_size[0], self.kernel_size[1])

        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                # import math
                # bound = math.sqrt(1 / (self.in_channels * self.kernel_size[0] * self.kernel_size[1]))
                # self.bias_value = np.random.uniform(-bound, bound, self.out_channels)
                self.bias_value = np.empty(self.out_channels)

            else:
                assert (bias.size == self.out_channels), f"The size of the given bias {bias.shape} does not correspond" \
                                                     f" to the size of output_channels {self.out_channels} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

        Hin = self.pre.shape[-2]
        Win = self.pre.shape[-1]
        if self.upscale is not None:
            assert isinstance(self.upscale, int)
            Win = Win*self.upscale
            Hin = Hin*self.upscale

        pool_Flag = False
        if 'avgpool' in self.synapse_name:
            pool_Flag = True
            self.pool_index = self.synapse_name.index('avgpool')

        if 'maxpool' in self.synapse_name:
            pool_Flag = True
            self.pool_index = self.synapse_name.index('maxpool')

        self.conv_index = self.synapse_name.index('conv')

        # if self.pool is not None:  # 池化
        if pool_Flag:
            self.pool_only = False
            self.pool_kernel_size = kwargs.get('pool_kernel_size', [2, 2])
            self.pool_stride = kwargs.get('pool_stride', 2)
            self.pool_padding = kwargs.get('pool_padding', 0)
            if isinstance(self.pool_kernel_size, int):
                self.pool_kernel_size = [self.pool_kernel_size] * 2
            if isinstance(self.pool_stride, int):
                self.pool_stride = [self.pool_stride] * 2
            if isinstance(self.pool_padding, int):
                self.pool_padding = [self.pool_padding] * 2

            self._constant_variables['pool_kernel_size[link]'] = self.pool_kernel_size
            self._constant_variables['pool_stride[link]'] = self.pool_stride
            self._constant_variables['pool_padding[link]'] = self.pool_padding

            if self.pool_index < self.conv_index:
                # pooling before conv
                self.pool_before = True
                Ho = int((Hin + 2 * self.pool_padding[0] - self.pool_kernel_size[0]) / self.pool_stride[0] + 1)
                Wo = int((Win + 2 * self.pool_padding[1] - self.pool_kernel_size[1]) / self.pool_stride[1] + 1)
                Ho = int((Ho + 2 * self.padding[0] - self.kernel_size[0]) / self.stride[0] + 1)
                Wo = int((Wo + 2 * self.padding[1] - self.kernel_size[1]) / self.stride[1] + 1)
            else:
                # conv before pooling
                self.pool_before = False
                Ho = int((Hin + 2 * self.padding[0] - self.kernel_size[0]) / self.stride[0] + 1)  # Ho = (Hin + 2 * padding[0] - kernel_size[0]) / stride[0] + 1
                Wo = int((Win + 2 * self.padding[1] - self.kernel_size[1]) / self.stride[1] + 1)  # Wo = (Win + 2 * padding[0] - kernel_size[1]) / stride[0] + 1
                # Ho = int(Ho / self.pool_kernel_size[0])
                # Wo = int(Wo / self.pool_kernel_size[1])
                Ho = int((Ho + 2 * self.pool_padding[0] - self.pool_kernel_size[0]) / self.pool_stride[0] + 1)  # Ho = (Hin + 2 * padding[0] - kernel_size[0]) / stride[0] + 1
                Wo = int((Wo + 2 * self.pool_padding[1] - self.pool_kernel_size[1]) / self.pool_stride[1] + 1)
        else:
            # without pooling
            Ho = int((Hin + 2 * self.padding[0] - self.kernel_size[0]) / self.stride[0] + 1)
            Wo = int((Win + 2 * self.padding[1] - self.kernel_size[1]) / self.stride[1] + 1)

        post_num = int(Ho * Wo * self.out_channels)

        if self.post.num == None:
            self.post.num = post_num
            self.post.shape = (self.out_channels, Ho, Wo)

        if self.post.num != None:
            if self.post.num != post_num:
                raise ValueError(
                    "The post_group num is not equal to the output num, cannot achieve the conv connection, "
                    "the output num is %d * %d * %d " % (self.out_channels, Ho, Wo))
            else:
                self.post.shape = (self.out_channels, Ho, Wo)

    def condition_check(self, pre_group, post_group):
        '''
        check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

        Args:
            pre_group(Groups): the neuron group which need to be connected in the pre.
            post_group(Groups): the neuron group which need to connect the pre_group in the post.

        Returns: flag

        '''

        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
Connection.register('conv', conv_connect)
Connection.register('conv_connection', conv_connect)

class ConvTranspose_connect(Connection):

    '''
    do the convolution connection.

    Args:
        pre(Assembly): the assembly which needs to be connected
        post(Assembly): the assembly which needs to connect the pre
        link_type(str): Conv
    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['conv'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['conv_transpose']
        super(ConvTranspose_connect, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)
        self.out_channels = kwargs.get('out_channels', None)
        self.in_channels = kwargs.get('in_channels', None)
        self.kernel_size = kwargs.get('kernel_size', [3, 3])
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.05)
        weight = kwargs.get('weight', None)

        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        self.mask = kwargs.get('mask', None)
        self.stride = _pair(kwargs.get('stride', 1))
        self.padding = _pair(kwargs.get('padding', 0))
        self.dilation = _pair(kwargs.get('dilation', 1))
        self.groups = kwargs.get('groups', 1)

        # self._variables = dict()
        self._constant_variables['stride[link]'] = self.stride
        self._constant_variables['padding[link]'] = self.padding
        self._constant_variables['dilation[link]'] = self.dilation
        self._constant_variables['groups[link]'] = self.groups

        if self.in_channels is None:
            from spaic.Neuron.Encoders import NullEncoder
            if len(self.pre.shape) == 1:
                raise ValueError('The shape %s of pre seems to be a flattened value. '
                                 'Give the correct shape with structure [channel, height, width] in the NeuronGroup '
                                 'class with the neuron_shape parameter' % self.pre.shape)
            elif isinstance(self.pre, NullEncoder):
                self.in_channels = self.pre.shape[2]
            else:
                self.in_channels = self.pre.shape[0]

        if self.out_channels is None:
            if len(self.post.shape) == 1:
                raise ValueError('The shape %s of post seems to be a flattened value.  '
                                 'Give the correct shape with structure [channel, height, width] in the NeuronGroup '
                                 'class with the neuron_shape parameter' % self.post.shape)
            else:
                self.out_channels = self.post.shape[0]

        self.shape = (self.out_channels, self.in_channels, self.kernel_size[0], self.kernel_size[1])

        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                # import math
                # bound = math.sqrt(1 / (self.in_channels * self.kernel_size[0] * self.kernel_size[1]))
                # self.bias_value = np.random.uniform(-bound, bound, self.out_channels)
                self.bias_value = np.empty(self.out_channels)

            else:
                assert (bias.size == self.out_channels), f"The size of the given bias {bias.shape} does not correspond" \
                                                     f" to the size of output_channels {self.out_channels} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

        Hin = self.pre.shape[-2]
        Win = self.pre.shape[-1]

        pool_Flag = False
        if 'avgpool' in self.synapse_name:
            pool_Flag = True
            self.pool_index = self.synapse_name.index('avgpool')

        if 'maxpool' in self.synapse_name:
            pool_Flag = True
            self.pool_index = self.synapse_name.index('maxpool')

        self.conv_index = self.synapse_name.index('conv')

        # if self.pool is not None:  # 池化
        if pool_Flag:
            raise NotImplementedError()
        else:
            # without pooling
            Ho = int(((Hin-1)*self.stride[0] - 2 * self.padding[0] - self.dilation[0]*(self.kernel_size[0]-1)) + 1)
            Wo = int(((Win-1)*self.stride[0] - 2 * self.padding[1] - self.dilation[1]*(self.kernel_size[1]-1)) + 1)

        post_num = int(Ho * Wo * self.out_channels)

        if self.post.num == None:
            self.post.num = post_num
            self.post.shape = (self.out_channels, Ho, Wo)

        if self.post.num != None:
            if self.post.num != post_num:
                raise ValueError(
                    "The post_group num is not equal to the output num, cannot achieve the conv connection, "
                    "the output num is %d * %d * %d " % (self.out_channels, Ho, Wo))
            else:
                self.post.shape = (self.out_channels, Ho, Wo)

    def condition_check(self, pre_group, post_group):
        '''
        check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

        Args:
            pre_group(Groups): the neuron group which need to be connected in the pre.
            post_group(Groups): the neuron group which need to connect the pre_group in the post.

        Returns: flag

        '''

        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
Connection.register('conv_transpose', conv_connect)




class ContinuousLocalConnection(Connection):
    '''
    Continuous local connection for 2D sheet of neurons that has 1 channel, and nearby neurons have similar
    connections.
    The conv parameter is fixed as stride=1, padding=0, dilation=1
    '''
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=None, max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn',
                 syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = []
        super(ContinuousLocalConnection, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name,
                                           syn_kwargs=syn_kwargs, **kwargs)
        self.w_std = kwargs.get('w_std', 0.5)
        self.w_mean = kwargs.get('w_mean', 0.05)
        self.w_min = kwargs.get('w_min', 0.0)
        self.w_max = kwargs.get('w_max', 0.1)
        self.kernel_size = _pair(kwargs.get('kernel_size', [3, 3]))
        self.clust_size = _pair(kwargs.get('clust_size', 1))
        self.dilation = _pair(kwargs.get('dilation', 1))
        # self.sigma_k = kwargs.get('sigma_k', 0.5*np.minimum(self.kernel_size[0], self.kernel_size[1]))
        self.sigma_c = kwargs.get('sigma_c', None)
        self.stride = _pair(kwargs.get('stride', 1))
        self.padding = _pair(kwargs.get('padding', 0))
        self.cutting = _pair(kwargs.get('cutting', 0))
        self._constant_variables['kernel_size[link]'] = self.kernel_size
        self._constant_variables['stride[link]'] = self.stride
        self._constant_variables['padding[link]'] = self.padding
        self._constant_variables['dilation[link]'] = self.dilation
        # the shape is fixed to (1, H, W), channel==1
        assert (len(self.pre.shape) == 3 or len(self.pre.shape)==4)and len(self.post.shape) == 3
        assert self.pre.shape[0] == 1 and self.post.shape[0] == 1
        if len(self.pre.shape) == 3:
            self.Hin = self.pre.shape[1]
            self.Win = self.pre.shape[2]
        else:
            self.Hin = self.pre.shape[2]
            self.Win = self.pre.shape[3]
        self.raw_Ho = int((self.Hin + 2 * self.padding[0] - self.dilation[0]*(self.kernel_size[0]-1) -1) // self.stride[0] + 1)
        self.raw_Wo = int((self.Win + 2 * self.padding[1] - self.dilation[1]*(self.kernel_size[1]-1) -1) // self.stride[1] + 1)
        self.Ho = self.clust_size[0]*self.raw_Ho - 2*self.cutting[0]
        self.Wo = self.clust_size[1]*self.raw_Wo - 2*self.cutting[1]

        post_num = self.Ho*self.Wo
        if self.post.num == None:
            self.post.num = post_num
            self.post.shape = (self.Ho, self.Wo)
        else:
            if not (self.post.shape[1] == self.Ho and self.post.shape[2] == self.Wo):
                raise ValueError(f"neuron shape {self.post.shape[1:]} != connection out shape {self.Ho, self.Wo}")
            assert post_num == self.post.num
        self.bias_flag = False

        # parameters for continuous local STDP
        self.use_STDP = kwargs.get('use_STDP', False)
        self.use_RSTDP = kwargs.get('use_RSTDP', False)
        self.A_plus = kwargs.get('A_plus', 1.0e-3)
        self.A_minus = kwargs.get('A_minus', 1.0e-4)
        self.tau_pre = kwargs.get('tau_pre', 10.0)
        self.tau_post = kwargs.get('tau_post', 100.0)
        self.homeo_add = kwargs.get('homeo_add', 1.0e-4)
        self.homeo_mult = kwargs.get('homeo_mult', 1.0e-3)
        self.reward_name = kwargs.get('reward_name', 'Output_Reward[updated]')
        self.trace_pre = None
        self.trace_post = None

    def init_weight(self):
        import torch
        from torch.nn.functional import conv2d
        clust_size2d = self.clust_size[0] * self.clust_size[1]
        assert clust_size2d >= 1

        # build the cluster gaussian filter
        # hc_size = 1 + 2 * int(2.0*self.clust_size[0] // 2)
        # wc_size = 1 + 2 * int(2.0*self.clust_size[1] // 2)
        # h_c = torch.arange(-(hc_size // 2), hc_size // 2 + 1).view(-1, 1).expand(-1, wc_size)
        # w_c = torch.arange(-(wc_size // 2), wc_size // 2 + 1).view(1, -1).expand(hc_size, -1)

        # print(torch.sum(torch.abs(clust_gauss)))
        # imshow(clust_gauss[0,0,:,:])
        # show()

        # initialize the weight
        H_input_size = self.Hin + 2*self.padding[0]
        W_input_szie = self.Win + 2*self.padding[1]
        H_mesh_size = self.clust_size[0] * (self.kernel_size[0] * self.dilation[0]) // self.stride[0]
        W_mesh_size = self.clust_size[1] * (self.kernel_size[1] * self.dilation[1]) // self.stride[1]
        H_output_size = self.raw_Ho*self.clust_size[0]
        W_output_size = self.raw_Wo*self.clust_size[1]
        if self.sigma_c is not None:
            if (self.kernel_size[0] * self.dilation[0])%self.stride[0] !=0 or \
                (self.kernel_size[1] * self.dilation[1])%self.stride[1] !=0:
                raise ValueError("Kernel_size must be integrate multiple of stride")


            self.sigma_c = 2*int(self.sigma_c//2) + 1
            weight = self.w_mean + self.w_std * torch.randn(H_input_size * W_input_szie, 1,
                                                            H_mesh_size + self.sigma_c - 1,
                                                            W_mesh_size + self.sigma_c - 1)

            h_f = torch.arange(-(self.sigma_c // 2), self.sigma_c // 2 + 1).view(-1, 1).expand(-1, self.sigma_c)
            w_f = torch.arange(-(self.sigma_c // 2), self.sigma_c // 2 + 1).view(1, -1).expand(self.sigma_c, -1)
            filt_r1 = (0.35*self.sigma_c)**2.0
            filt_r2 = (0.10*self.sigma_c)**2.0
            filter_DoG = (torch.exp(-(h_f ** 2 + w_f ** 2)/filt_r2)/filt_r2
                          - 1.0*torch.exp(-(h_f ** 2 + w_f ** 2)/filt_r1)/filt_r1).view(1, 1, self.sigma_c, self.sigma_c)
            self.clust_gauss = filter_DoG/torch.sum(filter_DoG**2)**0.5
            weight = conv2d(weight, filter_DoG)
            # print(torch.sum(self.clust_gauss**2)**0.5)
            # imshow(self.clust_gauss[0,0,:,:])
            # show()


            self.hw_indices = torch.zeros(self.kernel_size[0]*self.kernel_size[1], 1, H_output_size, W_output_size, dtype=torch.long)
            self.x_indices = torch.zeros(1, 1, H_output_size, W_output_size, dtype=torch.long)
            self.y_indices = torch.zeros(1, 1, H_output_size, W_output_size, dtype=torch.long)

            # x_indices, y_indices = torch.meshgrid([torch.arange(clust*raw_size), torch.arange(clust*raw_size)])

            for ii in range(H_output_size):
                for jj in range(W_output_size):
                    clust_stride_i = ii % H_mesh_size
                    kernel_i = ii // H_mesh_size
                    clust_stride_j = jj % W_mesh_size
                    kernel_j = jj// W_mesh_size
                    h_range = kernel_i * self.kernel_size[0] * self.dilation[0] + \
                              self.stride[0] * (clust_stride_i // self.clust_size[0]) +\
                              torch.arange(start=0, end=self.kernel_size[0] * self.dilation[0], step=self.dilation[0])
                    w_range = kernel_j * self.kernel_size[1] * self.dilation[1] + \
                              self.stride[1] * (clust_stride_j // self.clust_size[1]) +\
                              torch.arange(start=0, end=self.kernel_size[1] * self.dilation[1], step=self.dilation[1])
                    h_range, w_range = torch.meshgrid([h_range, w_range])
                    hw_range = h_range * W_input_szie + w_range
                    self.x_indices[0, 0, ii, jj] = clust_stride_i
                    self.y_indices[0, 0, ii, jj] = clust_stride_j
                    self.hw_indices[:, 0, ii, jj] = hw_range.view(-1)


            # imshow(self.x_indices[0, 0, :, :])
            # show()
            # imshow(self.y_indices[0, 0, :, :])
            # show()
            self.x_indices = self.x_indices[:, :, self.cutting[0]:H_output_size-self.cutting[0],
                             self.cutting[1]:W_output_size-self.cutting[1]]
            self.y_indices = self.y_indices[:, :, self.cutting[1]:W_output_size-self.cutting[1],
                             self.cutting[1]:W_output_size-self.cutting[1]]
            self.hw_indices = self.hw_indices[:, :, self.cutting[0]:H_output_size - self.cutting[0],
                             self.cutting[1]:W_output_size - self.cutting[1]]

            effective_weight = weight[self.hw_indices, 0, self.x_indices, self.y_indices]
            weight = torch.zeros_like(weight)
            weight[self.hw_indices, 0, self.x_indices, self.y_indices] = effective_weight

            self.shape = (H_input_size * W_input_szie, 1, H_mesh_size, W_mesh_size)
            self.H_input_size = H_input_size
            self.W_input_size = W_input_szie
            if (torch.max(self.hw_indices) >= self.shape[0]):
                print(f"wrong connection shape in {self.pre} to  {self.post}")

        else:
            weight = self.w_mean + self.w_std * torch.randn(self.kernel_size[0] * self.kernel_size[1], 1,
                                                            self.Ho, self.Wo)
            self.clust_gauss = None

            self.shape = (self.kernel_size[0]*self.kernel_size[1], 1,
                          self.Ho, self.Wo)

        # #normailze weight
        # weight = self.w_max*torch.sigmoid(self.w_mean + self.w_std*(weight - torch.mean(weight))/torch.std(weight))

        return weight
    def inti_conv_weight(self, weight):
        # from torch.nn.functional import conv2d
        if self.clust_gauss is not None:
            if weight.device != self.hw_indices.device:
                self.hw_indices = self.hw_indices.to(weight.device)
                self.x_indices = self.x_indices.to(weight.device)
                self.y_indices = self.y_indices.to(weight.device)
                self.clust_gauss = self.clust_gauss.to(weight.device)

            # con_weight = conv2d(weight, self.clust_gauss, padding=self.sigma_c // 2)
            # con_weight = torch.clamp(con_weight, self.w_min, self.w_max)
            # weight[:,:,:,:] = con_weight

            tmp_weight = weight[self.hw_indices, 0, self.x_indices, self.y_indices]
            tmp_mean = torch.mean(tmp_weight, dim=0, keepdim=True)
            tmp_weight = (tmp_weight + 0.2*(self.w_mean-tmp_mean))*(self.w_mean/tmp_mean)**0.5
            weight[self.hw_indices, 0, self.x_indices, self.y_indices] = tmp_weight
            weight.clamp_(self.w_min, self.w_max)
        # diff_mean = self.w_mean / (torch.mean(weight, dim=0, keepdim=True) + 1.0e-8)
        # weight.multiply_(diff_mean)

            # normailze weight
            # weight = self.w_max * torch.sigmoid(
            #     self.w_mean + 2*self.w_std * (weight - torch.mean(weight)) / torch.std(weight))
            # imshow(weight[100,0,:,:].detach().cpu())
            # show()

        return weight


    def local_connect_update(self, input, weight):
        import torch
        from torch.nn.functional import unfold


        # shape of input = (-1, 1, Hin, Win)
        # pad_input = pad(input, self.padding*2)
        unfold_input = unfold(input, self.kernel_size, padding=self.padding, stride=self.stride, dilation=self.dilation)
        # shape of unfold_input = (-1, self.kernel_size[0]*self.kernel_size[1], Ho*Wo)
        unfold_input = unfold_input.permute(1,0,2).reshape(
            self.kernel_size[0]*self.kernel_size[1], -1, self.raw_Ho, 1, self.raw_Wo, 1).expand(
            -1, -1, -1, self.clust_size[0], -1, self.clust_size[1]).reshape(self.kernel_size[0]*self.kernel_size[1], -1,
                                                                         self.raw_Ho*self.clust_size[0],
                                                                         self.raw_Wo*self.clust_size[1])
        unfold_input = unfold_input[:,:, self.cutting[0]:self.raw_Ho*self.clust_size[0]-self.cutting[0],
                       self.cutting[1]:self.raw_Wo*self.clust_size[1]-self.cutting[1]]
        if self.clust_gauss is not None:
            weight = weight[self.hw_indices, 0, self.x_indices, self.y_indices]
        output = torch.sum(unfold_input*weight, dim=0).unsqueeze(dim=1)
        return output

    def local_STDP_update(self, pre, post, weight):
        import torch
        from torch.nn.functional import unfold, conv2d
        if self.trace_pre is None:
            self.trace_pre = torch.zeros_like(pre)
            self.trace_post = torch.zeros_like(post)
            self.clust_gauss = self.clust_gauss.to(pre.device)

        # weight shape = (self.kernel_size[0]*self.kernel_size[1], 1,
        #                       self.Ho*self.clust_size[0], self.Wo*self.clust_size[1])
        # pre shape = (-1, 1, Hin, Win)
        # post shape = (-1, 1, Ho*clust_size[0], Wo*clust_size[1])
        self.trace_pre = self.beta_pre*self.trace_pre*pre.le(0.0) + pre
        self.trace_post = self.beta_post*self.trace_post*post.le(0.0) + post
        pre_reshape = torch.cat([pre, self.trace_pre], dim=0) # ( 2*-1, 1, Hin, Win)
        pre_reshape = unfold(pre_reshape, self.kernel_size, padding=self.padding, stride=self.stride, dilation=self.dilation)
        pre_reshape = pre_reshape.permute(1, 0, 2).reshape(
            self.kernel_size[0] * self.kernel_size[1], -1, self.Ho, 1, self.Wo, 1).expand(
            -1, -1, -1, self.clust_size[0], -1, self.clust_size[1]).reshape(self.kernel_size[0] * self.kernel_size[1],
                                                                            2, -1,
                                                                            self.Ho * self.clust_size[0],
                                                                            self.Wo * self.clust_size[1])
        trace_pre_reshape = pre_reshape[:,1,:, :, :]
        pre_reshape = pre_reshape[:,0,:, :, :]

        post_reshape = post.permute(1, 0, 2, 3).expand(self.kernel_size[0]*self.kernel_size[1], -1, -1, -1)
        trace_post_reshape = self.trace_post.permute(1, 0, 2, 3).expand(
            self.kernel_size[0]*self.kernel_size[1], -1, -1, -1)


        pre_post = trace_pre_reshape*post_reshape
        post_pre = trace_post_reshape*pre_reshape

        dw = torch.mean(self.A_plus*pre_post - self.A_minus*post_pre, dim=1, keepdim=True)
        if self.clust_gauss is not None:
            dw[self.hw_indices, 0, self.x_indices, self.y_indices] = dw
            dw = conv2d(dw, self.clust_gauss, padding=(self.hc_size//2, self.wc_size//2))


        # weight = torch.clamp(weight.add(dw),0,1)
        weight.add_(dw)
        weight.clamp_(0,0.1)
        diff_mean = self.w_mean - torch.mean(weight, dim=0, keepdim=True)
        weight.multiply_(torch.exp(self.homeo_mult * diff_mean))
        weight.add_(self.homeo_add*diff_mean)
        return weight

    def local_RSTDP_update(self,pre, post, reward, weight):
        import torch
        from torch.nn.functional import unfold, conv2d
        if self.trace_pre is None:
            self.trace_pre = torch.zeros_like(pre)
            self.trace_post = torch.zeros_like(post)
            if self.clust_gauss is not None:
                self.clust_gauss = self.clust_gauss.to(pre.device)


        self.trace_pre = self.beta_pre*self.trace_pre + (1-self.beta_pre)*pre
        self.trace_post = self.beta_post*self.trace_post + (1-self.beta_post)*post
        pre_reshape = torch.cat([pre, self.trace_pre], dim=0) # ( 2*-1, 1, Hin, Win)
        pre_reshape = unfold(pre_reshape, self.kernel_size, padding=self.padding, stride=self.stride, dilation=self.dilation)
        pre_reshape = pre_reshape.permute(1, 0, 2).reshape(
            self.kernel_size[0] * self.kernel_size[1], -1, self.raw_Ho, 1, self.raw_Wo, 1).expand(
            -1, -1, -1, self.clust_size[0], -1, self.clust_size[1]).reshape(self.kernel_size[0] * self.kernel_size[1],
                                                                            2, -1,
                                                                            self.raw_Ho * self.clust_size[0],
                                                                            self.raw_Wo * self.clust_size[1])
        trace_pre_reshape = pre_reshape[:,1,:, self.cutting[0]:self.raw_Ho*self.clust_size[0]-self.cutting[0],
                       self.cutting[1]:self.raw_Wo*self.clust_size[1]-self.cutting[1]]
        pre_reshape = pre_reshape[:,0,:, self.cutting[0]:self.raw_Ho*self.clust_size[0]-self.cutting[0],
                       self.cutting[1]:self.raw_Wo*self.clust_size[1]-self.cutting[1]]

        post_reshape = post.permute(1, 0, 2, 3).expand(self.kernel_size[0]*self.kernel_size[1], -1, -1, -1)
        trace_post_reshape = self.trace_post.permute(1, 0, 2, 3).expand(
            self.kernel_size[0]*self.kernel_size[1], -1, -1, -1)

        pre_post = trace_pre_reshape*post_reshape
        post_pre = trace_post_reshape*pre_reshape
        post_dw = self.A_plus*pre_post - self.A_minus*post_pre
        dw = torch.mean(reward.unsqueeze(0).unsqueeze(-1)*post_dw, dim=1, keepdim=True)
        if self.clust_gauss is not None:
            trans_dw = torch.zeros_like(weight)
            trans_dw[self.hw_indices, 0, self.x_indices, self.y_indices] = dw

            dw = conv2d(trans_dw, self.clust_gauss, padding=self.sigma_c // 2)
            # dw = dw * (dw.gt(0)*torch.exp(-0.001*self.w_max/(self.w_max-weight+1.0e-5))
            #            + dw.lt(0)*torch.exp(-0.001*self.w_max/(weight+1.0e-5)))
            dw = dw * (dw.gt(0)*torch.exp(-0.02*weight/self.w_max)
                       + dw.lt(0)*torch.exp(-0.02*(self.w_max-weight)/self.w_max))
            weight.add_(dw)



            # weight = conv2d(weight, self.clust_gauss, padding=self.sigma_c // 2)
            # for ii in range(100):
            #     subplot(2,1,1)
            #     imshow(weight[100,0,:,:].detach().cpu().numpy())
            #     weight = conv2d(weight, self.clust_gauss, padding=self.sigma_c // 2)
                # subplot(2, 1, 2)
            # imshow(weight[100, 0, :,:].detach().cpu().numpy())
            # show()
        else:
            dw = dw * (dw.gt(0) * torch.exp(-0.01 * self.w_max / (self.w_max - weight + 1.0e-5))
                       + dw.lt(0) * torch.exp(-0.01 * self.w_max / (weight + 1.0e-5)))
            weight.add_(dw)
            # weight = conv2d(weight, self.clust_gauss, padding=self.sigma_c // 2)

            # weight = torch.clamp(weight.add(dw),0,1)

            diff_mean = self.w_mean - torch.mean(weight, dim=0, keepdim=True)
            # weight.multiply_(diff_mean)
            weight.add_(self.homeo_add * diff_mean)
            weight.clamp_(self.w_min, self.w_max)
        # imshow(weight[100, 0, :,:].detach().cpu().numpy())
        # show()
        return weight




    def build(self, backend):
        from spaic.Network.BaseModule import Op
        self.weight = self.init_weight()
        self._variables['weight[link]'] = self.weight
        super(ContinuousLocalConnection, self).build(backend)
        update_op = Op(owner=self)
        update_op.func_name = self.local_connect_update
        update_op.output = self.add_conn_label(self.post_var_name + '[post]')
        update_op.input = [self.add_conn_label(self.pre_var_name + '[pre]'), self.add_conn_label('weight[link]')]
        backend.add_operation(update_op)
        self._operations.append(update_op)
        weightConv_op = Op(owner=self)
        weightConv_op.func_name = self.inti_conv_weight
        weightConv_op.output = self.add_conn_label('weight[link]')
        weightConv_op.input = self.add_conn_label('weight[link]')
        self._init_operations.append(weightConv_op)
        backend.register_initial(weightConv_op)

        if self.use_STDP:
            self.beta_pre = np.exp(-backend.dt/self.tau_pre)
            self.beta_post = np.exp(-backend.dt/self.tau_post)
            STDP_op = Op(owner=self)
            STDP_op.func_name = self.local_STDP_update
            STDP_op.output = self.add_conn_label('weight[link]')
            STDP_op.input = [self.add_conn_label(self.pre_var_name + '[pre]'),
                               self.add_conn_label('O[post]'),
                               self.add_conn_label('weight[link]')]
            backend.add_operation(STDP_op)
            self._operations.append(STDP_op)
        elif self.use_RSTDP:
            self.beta_pre = np.exp(-backend.dt/self.tau_pre)
            self.beta_post = np.exp(-backend.dt/self.tau_post)
            RSTDP_op = Op(owner=self)
            RSTDP_op.func_name = self.local_RSTDP_update
            RSTDP_op.output = self.add_conn_label('weight[link]')
            RSTDP_op.input = [self.add_conn_label(self.pre_var_name + '[pre]'),
                               self.add_conn_label('O[post]'),
                               self.reward_name,
                               self.add_conn_label('weight[link]')]
            backend.add_operation(RSTDP_op)
            self._operations.append(RSTDP_op)

Connection.register('continuous_local', ContinuousLocalConnection)


class ComplexAttentionConnection(Connection):
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=None, max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn',
                 syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = []
        super(ComplexAttentionConnection, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name,
                                           syn_kwargs=syn_kwargs, **kwargs)

        self.group_num = kwargs.get('group_num', 1) # multi_head number
        self.d_k = kwargs.get('d_k', 3)
        self.sqrt_dk = (self.d_k)**0.5
        self.w_std = kwargs.get('w_std', 0.5)
        self.w_mean = kwargs.get('w_mean', 0.05)
        assert self.post.num % self.group_num == 0
        self._variables['W_v[link]'] = torch.empty(self.pre_num, self.post_num) # w_v
        self._variables['W_k[link]'] = torch.empty(self.group_num, self.pre_num, self.d_k)
        self._variables['W_q[link]'] = torch.empty(self.group_num, self.pre_num, self.d_k*self.post_num // self.group_num)
        torch.nn.init.normal_(self._variables['W_v[link]'], self.w_mean, self.w_std)
        torch.nn.init.xavier_normal_(self._variables['W_k[link]'])
        torch.nn.init.xavier_normal_(self._variables['W_q[link]'])
        self._variables['Key[link]'] = torch.zeros((1, self.group_num, self.pre_num, self.d_k))
        self._variables['Query[link]'] = torch.zeros((1, self.group_num, self.post_num//self.group_num, self.d_k))
        self._variables['tau_m[link]'] = kwargs.get('tau_m', 10.0)
        self._variables['tau_r[link]'] = kwargs.get('tau_r', 5.0)
        self._variables['beta[link]'] = kwargs.get('beta', 1.0)



    def update(self, input0, Key, Query, beta, W_v, W_k, W_q):
        # input.shape = (batch, pre_neuron_num)
        # Key.shape = (batch, group_num, pre_neuron_num, d_k)
        # W_k.shape = (group_num, pre_neuron_num, d_k)
        # Query.shape = (batch, group_num, post_neuron_num//group_num, d_k)
        # W_q.shape = (group_num, pre_neuron_num, post_neuron_num//group_num * d_k)
        # Value.shape = (batch, group_num, pre_neuron_num, post_neuron_num//group_num)
        # W_v.shape = (group_num, pre_neuron_num, post_neuron_num//group_num)
        if input0.dtype.is_complex:
            input = beta**input0.imag * (1.0j*input0.real)
        else:
            input = input0*1.0j
        Key = beta*Key + input.view(-1, 1, self.pre_num, 1)*W_k.unsqueeze(0)
        Query = beta*Query + torch.complex(torch.matmul(input.view(-1, 1, 1, self.pre.num).real,
                                          W_q).view(-1, self.group_num, self.post.num//self.group_num, self.d_k),
                                           torch.matmul(input.view(-1, 1, 1, self.pre.num).imag,W_q).
                                           view(-1, self.group_num, self.post.num // self.group_num,self.d_k))

        Z = torch.softmax(torch.matmul(Query.real, Key.transpose(3,2).real)/self.sqrt_dk, dim=-2).view(-1, self.pre_num, self.post_num) # shape=(batch, group_num, 1, pre_neuron_num)
        Value = input.view(-1, self.pre.num, 1)*W_v
        Isyn = torch.sum(Z*Value, dim=-2).view(-1, self.post.num) # shape=(batch, post_neuron_num)

        return Key, Query, Isyn

    def init_beta(self, tau_m, tau_r, dt):
        beta = torch.exp(-dt/tau_m -2*np.pi*dt*1.0j/tau_r)
        return beta

    def build(self, backend):
        super(ComplexAttentionConnection, self).build(backend)
        from spaic.Network.BaseModule import Op
        Wv_name = self.add_conn_label('W_v[link]')
        Wk_name = self.add_conn_label('W_k[link]')
        Wq_name = self.add_conn_label('W_q[link]')
        beta_name = self.add_conn_label('beta[link]')
        Key_name = self.add_conn_label('Key[link]')
        Query_name = self.add_conn_label('Query[link]')
        update_op = Op(owner=self)
        update_op.func_name = self.update
        update_op.output = [Key_name, Query_name, self.add_conn_label(self.post_var_name + '[post]')]
        update_op.input = [self.add_conn_label(self.pre_var_name + '[pre]'), Key_name, Query_name,
                           beta_name, Wv_name, Wk_name, Wq_name]
        backend.add_operation(update_op)
        self._operations.append(update_op)

        init_op = Op(owner=self)
        init_op.func_name = self.init_beta
        init_op.output = self.add_conn_label('beta[link]')
        init_op.input = [self.add_conn_label('tau_m[link]'), self.add_conn_label('tau_r[link]'), '[dt]']
        self._init_operations.append(init_op)
        backend.register_initial(init_op)


Connection.register('complex_attention', ComplexAttentionConnection)







class pool_connect(Connection):

    '''
    do the pooling operation on spiking trains.

    Args:
        pre(Assembly): the assembly which needs to be connected
        post(Assembly): the assembly which needs to connect the pre
        link_type(str): Conv
    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['maxpool'], max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['maxpool']
        super(pool_connect, self).__init__(pre=pre, post=post, name=name,
                                           link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                           sparse_with_mask=sparse_with_mask,
                                           pre_var_name=pre_var_name, post_var_name=post_var_name, syn_kwargs=syn_kwargs, **kwargs)

        self.pool_kernel_size = kwargs.get('pool_kernel_size', [2, 2])
        self.pool_stride = kwargs.get('pool_stride', 2)
        self.pool_padding = kwargs.get('pool_padding', 0)
        if isinstance(self.pool_kernel_size, int):
            self.pool_kernel_size = [self.pool_kernel_size] * 2
        if isinstance(self.pool_stride, int):
            self.pool_stride = [self.pool_stride] * 2
        if isinstance(self.pool_padding, int):
            self.pool_padding = [self.pool_padding] * 2

        self._constant_variables['pool_kernel_size[link]'] = self.pool_kernel_size
        self._constant_variables['pool_stride[link]'] = self.pool_stride
        self._constant_variables['pool_padding[link]'] = self.pool_padding
        self.pool_only = True


    def condition_check(self, pre_group, post_group):
        '''
        check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

        Args:
            pre_group(Groups): the neuron group which need to be connected in the pre.
            post_group(Groups): the neuron group which need to connect the pre_group in the post.

        Returns: flag

        '''

        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag

Connection.register('pool', pool_connect)
Connection.register('pool_connection', pool_connect)


class sparse_connect_sparse(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(sparse_connect_sparse, self).__init__(pre=pre, post=post, name=name,
                                                    link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                    sparse_with_mask=sparse_with_mask,
                                                    pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                    syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.w_std = kwargs.get('w_std', 0.05)
        self.density = kwargs.get('density', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', False)
        self.is_sparse = kwargs.get('is_sparse', True)
        # self._variables = dict()

        if weight is None:
            # Connection weight
            sparse_matrix = self.w_std * sp.rand(*self.shape, density=self.density, format='csr')
            self.weight = sparse_matrix.toarray()
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                if self.w_init == 'sparse':
                    self.weight = np.empty(self.shape)
                else:
                    raise ValueError('Sparse implementation of sparse_connection_sparse connection do not support '
                                     'the %s initialization method', self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('sparse_sparse', sparse_connect_sparse)
Connection.register('sparse_connection_sparse', sparse_connect_sparse)


class sparse_connect_mask(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(sparse_connect_mask, self).__init__(pre=pre, post=post, name=name,
                                                  link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                  sparse_with_mask=sparse_with_mask,
                                                  pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                  syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.density = kwargs.get('density', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        self.is_int = kwargs.get('is_int', False)
        self.kwargs = kwargs
        # self._variables = dict()

        if weight is None:
            # Connection weight
            # sparse_matrix = self.w_std * sp.rand(*self.shape, density=self.density, format='csr')
            # self.weight = sparse_matrix.toarray()
            # self.weight[self.weight.nonzero()] = self.weight[self.weight.nonzero()] + self.w_mean

            ########################### this takes a lot of time################################
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
            if self.w_max is not None or self.w_min is not None:
                # self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
                self.weight = np.clip(self.w_std * np.random.randn(*self.shape) + self.w_mean, a_min=self.w_min, a_max=self.w_max)
            else:
                self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
            sparse_mask = np.less(np.random.rand(*self.shape), self.density)
            self.weight = np.multiply(self.weight, sparse_mask)
            ###################################################################################
            if self.is_int:
                self.weight = np.round(self.weight)
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError('Dense implementation of sparse_connection_mask do not support the %s initialization'
                                 ' method', self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('sparse', sparse_connect_mask)
Connection.register('sparse_connection', sparse_connect_mask)


class random_connect_sparse(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(random_connect_sparse, self).__init__(pre=pre, post=post, name=name,
                                                    link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                    sparse_with_mask=sparse_with_mask,
                                                    pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                    syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.probability = kwargs.get('probability', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', False)
        self.is_sparse = kwargs.get('is_sparse', True)
        # self._variables = dict()

        if weight is None:
            # Link_parameters
            prob_weight = np.random.rand(*self.shape)
            diag_index = np.arange(min([self.pre_num, self.post_num]))
            prob_weight[diag_index, diag_index] = 1
            index = (prob_weight < self.probability)
            # Connection weight
            self.weight = np.zeros(self.shape)
            self.weight[index] = prob_weight[index]
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                if self.w_init == 'sparse':
                    self.weight = np.empty(self.shape)
                else:
                    raise ValueError('Sparse implementation of random_connect_sparse connection do not support '
                                     'the %s initialization method', self.w_init)
                self.weight = np.random.randn(*self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('random_connection_sparse', random_connect_sparse)



class random_connect_mask(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(random_connect_mask, self).__init__(pre=pre, post=post, name=name,
                                                  link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                                  sparse_with_mask=sparse_with_mask,
                                                  pre_var_name=pre_var_name, post_var_name=post_var_name,
                                                  syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        self.probability = kwargs.get('probability', 0.1)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        # self.init = kwargs.get('init', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        # self._variables = dict()

        if weight is None:
            # Link_parameters
            prob_weight = np.random.rand(*self.shape)
            diag_index = np.arange(min([self.pre_num, self.post_num]))
            prob_weight[diag_index, diag_index] = 1
            index = (prob_weight < self.probability)
            # Connection weight
            self.weight = np.zeros(self.shape)
            self.weight[index] = prob_weight[index]
        else:
            if isinstance(weight, BaseInitializer):
                raise ValueError(
                    'Dense implementation of random_connect_mask do not support the %s initialization method',
                    self.w_init)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight

        bias = kwargs.get('bias', None)
        if bias is None:
            self.bias_flag = False
        else:
            self.bias_flag = True
            if isinstance(bias, BaseInitializer):
                self.b_init, self.b_init_param = self.decode_initializer(bias)
                self.bias_value = np.empty(self.post_num)

            else:
                assert (bias.size == self.post_num), f"The size of the given bias {bias.shape} does not correspond" \
                                                         f" to the size of post_num {self.post_num} "
                self.bias_value = bias

            self._variables['bias[link]'] = self.bias_value

    def condition_check(self, pre_group, post_group):
        flag = False
        pre_type = pre_group.type
        post_type = post_group.type
        if pre_type == post_type:
            flag = True
        return flag
        pass
Connection.register('random', random_connect_mask)
Connection.register('random_connection', random_connect_mask)


class NullConnection(Connection):

    '''
    each neuron in the first layer is connected to each neuron in the second layer.

    Args:
        pre(Assembly): The assembly which needs to be connected
        post(Assembly): The assembly which needs to connect the pre
        link_type(str): full

    Methods:
        unit_connect: define the basic connection information and add them to the connection_information.
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.

    '''

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(NullConnection, self).__init__(pre=pre, post=post, name=name,
                                             link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                             sparse_with_mask=sparse_with_mask,
                                             pre_var_name=pre_var_name, post_var_name=post_var_name,
                                             syn_kwargs=syn_kwargs, **kwargs)

        pass

Connection.register('null', NullConnection)

class Spike_to_Mass(Connection):
    '''
    Connect pre of spiking neurons to post of Neural Mass model
    '''
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(Spike_to_Mass, self).__init__(pre=pre, post=post, name=name,
                                            link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                            sparse_with_mask=sparse_with_mask,
                                            pre_var_name=pre_var_name, post_var_name=post_var_name,
                                            syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        bias = kwargs.get('bias', None)
        self.tau_syn = kwargs.get('tau_syn', 6.0)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.shape = (post.num, 1)
        self.synapse_class = []


        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight
        self._variables['beta[link]'] = np.exp(-0.1/self.tau_syn)
        self._variables['rate[link]'] = 0.0
        self._operations.append(['rate[link]', self.spike_to_rate, [pre_var_name+'[pre]', 'rate[link]', 'beta[link]']])
        self._operations.append([self.post_var_name+'[post]', 'mat_mult_weight', 'rate[link][updated]', 'weight[link]'])


    def spike_to_rate(self, spikes, last_rate, beta):
        rate = beta*last_rate + torch.mean(spikes, dim=1, keepdim=True)
        return rate

Connection.register('spike2mass', Spike_to_Mass)

class Mass_to_Spike(Connection):
    '''
    Connect pre of spiking neurons to post of Neural Mass model
    '''
    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(Mass_to_Spike, self).__init__(pre=pre, post=post, name=name,
                                            link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                            sparse_with_mask=sparse_with_mask,
                                            pre_var_name=pre_var_name, post_var_name=post_var_name,
                                            syn_kwargs=syn_kwargs, **kwargs)
        weight = kwargs.get('weight', None)
        bias = kwargs.get('bias', None)
        self.tau_syn = kwargs.get('tau_syn', 6.0)
        self.w_std = kwargs.get('w_std', 0.05)
        self.w_mean = kwargs.get('w_mean', 0.005)
        self.w_max = kwargs.get('w_max', None)
        self.w_min = kwargs.get('w_min', None)
        self.is_parameter = kwargs.get('is_parameter', True)
        self.shape = (post.num, pre.num)
        self.synapse_class = []
        self.unit_conversion = kwargs.get('unit_conversion', 1.0)


        if weight is None:
            # Connection weight
            self.weight = self.w_std * np.random.randn(*self.shape) + self.w_mean
        else:
            if isinstance(weight, BaseInitializer):
                self.w_init, self.w_init_param = self.decode_initializer(weight)
                self.weight = np.empty(self.shape)
            else:
                assert (weight.shape == self.shape), f"The size of the given weight {weight.shape} does not correspond" \
                                                     f" to the size of synaptic matrix {self.shape} "
                self.weight = weight

        self._variables['weight[link]'] = self.weight
        self._variables['beta[link]'] = np.exp(-0.1/self.tau_syn)
        self._variables['rate[link]'] = 0.0

        self._operations.append([post_var_name+'[post]', self.rate_to_Isyn, [pre_var_name+'[pre]', 'weight[link]']])



    def rate_to_Isyn(self, rate, weight):
        rate = rate.view(rate.shape[0], 1, self.pre.num).expand(rate.shape[0],
                                                                self.post.num, self.pre.num)
        spikes = torch.rand_like(rate).le(rate * self.unit_conversion * self.dt).type(self._backend.data_type)
        Isyn = torch.sum(spikes*weight, dim=-1)
        return Isyn
Connection.register('mass2spike', Mass_to_Spike)

class DistDepd_connect(Connection):

    def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=['basic'], max_delay=0, sparse_with_mask=False, pre_var_name='O',
                 post_var_name='Isyn', syn_kwargs=None, **kwargs):
        if syn_type is None: syn_type = ['basic']
        super(DistDepd_connect, self).__init__(pre=pre, post=post, name=name,
                                               link_type=link_type, syn_type=syn_type, max_delay=max_delay,
                                               sparse_with_mask=sparse_with_mask,
                                               pre_var_name=pre_var_name, post_var_name=post_var_name,
                                               syn_kwargs=syn_kwargs, **kwargs)
        self.distance_weight_function = kwargs.get('distance_weight_function', None)
        self.zero_self = kwargs.get('zero_self', False)
        if self.distance_weight_function is None:
            self.distance_weight_function = self.default_dist_weight_function
        self.dist_function = kwargs.get('dist_function', 'euclidean')
        if self.dist_function == 'euclidean':
            self.dist_function = self.euclidean_dist_function
        elif self.dist_function == 'circular':
            self.dist_function = self.circular_dist_function

        self.dist_a = kwargs.get('dist_a', 0.2)
        self.dist_b = kwargs.get('dist_b', 0.4)
        self.w_amp = kwargs.get('w_amp', -0.1)
        self.pos_range = kwargs.get('pos_range', 1.0)
        self.p = kwargs.get('p', 1.0)

        # building the weight
        pre_num = pre.num
        post_num = post.num
        assert len(pre.position) > 0
        assert len(post.position) > 0

        post_pos = np.expand_dims(post.position, axis=1)
        pre_pos = np.expand_dims(pre.position, axis=0)
        weight = self.distance_weight_function(self.dist_function(pre_pos, post_pos))
        if self.p < 1.0:
            rand_mask = torch.rand_like(weight).lt(self.p)
            weight = weight*rand_mask
        self._variables['weight[link]'] = weight





    def unit_connect(self, pre_group, post_group):
        # The number of neurons in neuron group
        pre_num = pre_group.num
        post_num = post_group.num
        assert len(pre_group.position) > 0
        assert len(post_group.position) > 0
        link_num = pre_num * post_num
        shape = (post_num, pre_num)
        weight = np.zeros(shape)
        post_pos = np.expand_dims(post_group.position,  axis=1)
        pre_pos = np.expand_dims(pre_group.position, axis=0)
        weight = self.distance_weight_function(self.dist_function(pre_pos, post_pos))
        # from matplotlib import pyplot as plt
        # plt.imshow(weight)
        # plt.show()

        # The name for backend variables
        input_name = self.get_input_name(pre_group, post_group)
        weight_name = self.get_weight_name(pre_group, post_group)
        target_name = self.get_target_name(post_group)

        # The backend variable
        var_code = (weight_name, shape, weight, True, False) # (var_name, shape, value, is_parameter, is_sparse, init)
        op_code = [target_name, 'mat_mult', input_name, weight_name]
        connection_information = (pre_group, post_group, link_num, var_code, op_code)
        self.unit_connections.append(connection_information)

    def circular_dist_function(self, pre_pos, post_pos):
        if not isinstance(pre_pos, torch.Tensor):
            pre_pos = torch.tensor(pre_pos)
        if not  isinstance(post_pos, torch.Tensor):
            post_pos = torch.tensor(post_pos)

        z = torch.maximum(pre_pos, post_pos)
        k = torch.minimum(pre_pos, post_pos)
        dist = torch.minimum(z - k, self.pos_range + k - z)
        dist = torch.norm(dist, p=2, dim=-1)
        return dist

    def euclidean_dist_function(self, pre_pos, post_pos):
        if isinstance(pre_pos, torch.Tensor) and isinstance(post_pos, torch.Tensor):
            diff = pre_pos - post_pos
        else:
            diff = torch.tensor(pre_pos-post_pos)
        dist = torch.norm(diff, p=2, dim=-1)
        return dist

    def default_dist_weight_function(self, dist):
        weights = self.w_amp * (torch.exp(-dist / self.dist_a)/self.dist_a - 0.5*torch.exp(-dist / self.dist_b)/self.dist_b)
        if self.zero_self:
            weights = weights * (dist!=0).type(weights.dtype)
        # import matplotlib.pyplot as plt
        # plt.imshow(weights, aspect='auto')
        # plt.show()
        return weights
Connection.register('dist_depd', DistDepd_connect)

# class reconnect(Connection):
#     def __init__(self, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'), policies=[],
#                  max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn', syn_kwargs=None, **kwargs):
#         super(reconnect, self).__init__(pre=pre, post=post, name=name, link_type=link_type,
#                                              policies=policies, max_delay=max_delay, sparse_with_mask=sparse_with_mask, pre_var_name=pre_var_name, post_var_name=post_var_name,syn_kwargs=syn_kwargs, **kwargs)
#     def unit_connect(self, pre_group, post_group):
#         pass
#
#     def condition_check(self, pre_group, post_group):
#         pass



# -*- coding: utf-8 -*-
"""
Created on 2020/8/11
@project: SPAIC
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""

from .STCA_Learner import STCA
from .STBP_Learner import STBP
from .RSTDP import RSTDP, RSTDPET
# from .PSD_Learner import PSD

from .Rate_Modulation import Rate_Modulate
from .Backprop_RSTDP import Backprop_RSTDP

from .STDP_Learner import nearest_online_STDP, full_online_STDP
from .Conv_RSTDP import Conv2d_RSTDP
from .Conv_STDP import Conv2d_STDP
# from .Tempotron_stm import Tempotron_stm
from .Tempotron_Learner import Tempotron



# -*- coding: utf-8 -*-
"""
Created on 2021/5/10
@project: SPAIC
@filename: ConnectionPolicy
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""

from spaic.Network.Assembly import Assembly
from spaic.Network.Topology import ConnectPolicy, Projection
# from spaic.Neuron.Neuron import NeuronGroup
# from spaic.Neuron.Node import Node



# class ConnectInformation():
#     _con_count = 0
#
#
#     def __init__(self, pre: Assembly, post: Assembly, level=0):
#         super(ConnectInformation, self).__init__()
#         self.pre = pre
#         self.post = post
#         self.abstract_level = 0
#         self.my_level = level
#         self.super = None
#         if (self.pre.id is not None) and (self.post.id is not None):
#             self.key = self.pre.id + "->" + self.post.id
#         else:
#             self.key = 'default' + str(ConnectInformation._con_count)
#         ConnectInformation._con_count += 1
#
#         self._link_num = 0
#         self._sub_unit_connections = dict()
#         self._sub_assb_connections = dict()
#         self._leaf_connections = {self.key: self}
#
#         if pre._is_terminal and post._is_terminal:
#             self.is_unit = True
#         else:
#             self.is_unit = False
#
#
#     def homologous(self, other):
#         if (other.pre is self.pre) and (other.post is self.post):
#             return True
#         else:
#             return False
#
#     def is_empty(self):
#         if self.is_unit:
#             return False
#         elif len(self._sub_unit_connections) + len(self._sub_assb_connections) == 0:
#             return True
#         else:
#             return False
#
#     def __and__(self, other):
#         if not self.homologous(other):
#             raise ValueError("can't do & operation for nonhomologous connections")
#         if self.is_empty():
#             new_connect = other
#         elif other.is_empty():
#             new_connect = self
#         else:
#             new_connect = ConnectInformation(self.pre, self.post)
#             # unit connections
#             key1 = set(self._sub_unit_connections.keys())
#             key2 = set(other._sub_unit_connections.keys())
#             unite_keys = key1.intersection(key2)
#             for key in unite_keys:
#                 new_connect._sub_unit_connections[key] = self._sub_unit_connections[key]
#                 new_connect._sub_unit_connections[key].super = new_connect
#
#             #assb connections
#             key1 = set(self._sub_assb_connections.keys())
#             key2 = set(other._sub_assb_connections.keys())
#             unite_keys = key1.intersection(key2)
#             for key in unite_keys:
#                 new_connect._sub_assb_connections[key] = self._sub_assb_connections[key] & other._sub_assb_connections[key]
#                 new_connect._sub_assb_connections[key].super = new_connect
#
#         return new_connect
#
#
#     def add_connection(self, con):
#         con.super = self
#         key = con.key
#         if not (con.pre in self.pre and con.post in self.post):
#             raise ValueError("the sub connection is not belong to this connection group (pre and post is not a member of the connected Assemblies)")
#
#         if con.is_unit:
#             self._sub_unit_connections[key] = con
#         else:
#             self._sub_assb_connections[key] = con
#
#         # if len(self._level_connections) >= 2:
#         #     self._level_connections[1].append(con)
#         # else:
#         #     self._level_connections.append([con])
#
#         # new_level = self.my_level + 1
#         # if len(self.super._level_connections) >= new_level+1:
#         #     self.super._level_connections[new_level].append(con)
#         # else:
#         #     self.super._level_connections.append([con])
#
#
#     def expand_connection(self, to_level=-1):
#         assert isinstance(to_level, int), ValueError("level is not int")
#         if self.is_unit:
#             return self._leaf_connections.values()
#         else:
#             new_leaf_connections = dict()
#             assb_connections = self._leaf_connections.values()
#             while(len(assb_connections)>0):
#                 if to_level >= 0 and to_level<=self.abstract_level:
#                     break
#                 self.abstract_level += 1
#                 new_assb_connections = []
#                 for con in assb_connections:
#                     if con.is_unit:
#                         new_leaf_connections[con.key] = con
#                     else:
#                         pre_groups = con.pre.get_groups(recursive=False)
#                         post_groups = con.post.get_groups(recursive=False)
#                         for pre in pre_groups:
#                             for post in post_groups:
#                                 new_con = ConnectInformation(pre, post, self.abstract_level)
#                                 con.add_connection(new_con)
#                                 if new_con.is_unit:
#                                     new_leaf_connections[new_con.key] = new_con
#                                 else:
#                                     new_assb_connections.append(new_con)
#
#                 assb_connections = new_assb_connections
#
#             for con in assb_connections:
#                 #对于超过abstract_level的connection，都认为是leaf_connection
#                 new_leaf_connections[con.key] = con
#             self._leaf_connections = new_leaf_connections
#             return list(new_leaf_connections.values())
#
#
#     def del_connection(self):
#         if self.super is not None:
#             if self.is_unit:
#                 self.super._sub_unit_connections.pop(self.key)
#             else:
#                 self.super._sub_assb_connections.pop(self.key)
#             top_leaf = self.top._leaf_connections
#             if self.key in top_leaf:
#                 top_leaf.pop(self.key)
#             if self.super.is_empty():
#                 self.super.del_connection()
#         else:
#             self.is_unit = False
#
#
#     # def add_var_code(self, var_code):
#     #     if self.is_unit:
#     #         self._var_codes.append(var_code)
#     #     else:
#     #         raise ValueError("can't add var_code for non-unit connections")
#     #
#     # def add_op_code(self, op_code):
#     #     if self.is_unit:
#     #         self._op_codes.append(op_code)
#     #     else:
#     #         raise ValueError("can't add op_code for non-unit connections")
#
#     def set_link_num(self, link_num):
#         if self.is_unit:
#             self._link_num = link_num
#             if self.super is not None:
#                 self.super.add_link_num(link_num)
#         else:
#             raise ValueError("can't set link_num for non-unit connections")
#
#     def add_link_num(self, link_num):
#         self._link_num += link_num
#         if self.super is not None:
#             self.super.add_link_num(link_num)
#
#
#     @property
#     def top(self):
#         if self.super is None:
#             return self
#         else:
#             return self.super.top
#
#     @property
#     def link_num(self):
#         return self._link_num
#
#     @property
#     def sub_connections(self):
#         if self.is_unit:
#             raise ValueError("no sub_connections for unit connections")
#         else:
#             sub_connections = []
#             sub_connections.extend(self._sub_assb_connections.values())
#             sub_connections.extend(self._sub_unit_connections.values())
#             return sub_connections
#
#     @property
#     def leaf_connections(self):
#         return list(self._leaf_connections.values())
#
#     @property
#     def sub_unit_connections(self):
#         if self.is_unit:
#             raise ValueError("no sub_connections for unit connections")
#         else:
#             return list(self._sub_unit_connections.values())
#
#     @property
#     def sub_assb_connections(self):
#         if self.is_unit:
#             raise ValueError("no sub assb_connections for unit connections")
#         else:
#             return list(self._sub_assb_connections.values())
#
#     @property
#     def all_unit_connections(self):
#         if self.is_unit:
#             return [self]
#         else:
#             unit_connections = []
#             unit_connections.extend(self._sub_unit_connections.values())
#             for con in self._sub_assb_connections.values():
#                 unit_connections.extend(con.all_unit_connections)
#             return unit_connections
#
#     @property
#     def all_assb_connections(self):
#         if self.is_unit:
#             raise ValueError("no sub_connections for unit connections")
#         else:
#             assb_connections = []
#             assb_connections.extend(self._sub_assb_connections.values())
#             for con in self._sub_assb_connections.values():
#                 assb_connections.extend(con.all_assb_connections)
#             return assb_connections




class IncludedTypePolicy(ConnectPolicy):

    def __init__(self, pre_types=None, post_types=None, level=-1):
        super(IncludedTypePolicy, self).__init__(level=level)
        self.name = 'Included_policy'
        if isinstance(pre_types, list):
            self.pre_types = pre_types
        elif pre_types is not None:
            self.pre_types = [pre_types]
        else:
            self.pre_types = None

        if isinstance(post_types, list):
            self.post_types = post_types
        elif post_types is not None:
            self.post_types = [post_types]
        else:
            self.post_types = None


    def checked_connection(self, new_connection: Projection):

        leaf_connections = new_connection.expand_connection(self.level)
        if self.post_types is not None:
            self.post_types = set(self.post_types)
            for con in leaf_connections:
                fit_type = self.post_types.intersection(con.post.type)
                if not fit_type:
                    con.del_connection()

        if self.pre_types is not None:
            leaf_connections = new_connection.leaf_connections
            self.pre_types = set(self.pre_types)
            for con in leaf_connections:
                fit_type = self.pre_types.intersection(con.pre.type)
                if not fit_type:
                    con.del_connection()
                    
        return new_connection


class ExcludedTypePolicy(ConnectPolicy):

    def __init__(self, pre_types=None, post_types=None, level=-1):
        super(ExcludedTypePolicy, self).__init__(level=level)
        self.name = 'Excluded_policy'
        if isinstance(pre_types, list):
            self.pre_types = pre_types
        elif pre_types is not None:
            self.pre_types = [pre_types]
        else:
            self.pre_types = None

        if isinstance(post_types, list):
            self.post_types = post_types
        elif post_types is not None:
            self.post_types = [post_types]
        else:
            self.post_types = None

    def checked_connection(self, new_connection: Projection):

        leaf_connections = new_connection.expand_connection(self.level)
        if isinstance(leaf_connections, dict):
            leaf_connections = leaf_connections.values()

        if self.post_types is not None:
            self.post_types = set(self.post_types)
            for con in leaf_connections:
                fit_type = self.post_types.intersection(con.post.type)
                if fit_type:
                    con.del_connection()

        if self.pre_types is not None:
            leaf_connections = new_connection._leaf_connections
            self.pre_types = set(self.pre_types)
            for con in leaf_connections:
                fit_type = self.pre_types.intersection(con.pre.type)
                if fit_type:
                    con.del_connection()
        return new_connection


class IndexConnectPolicy(ConnectPolicy):

    def __init__(self, pre_indexs=None, post_indexs=None, level=-1):
        super(IndexConnectPolicy, self).__init__(level=level)
        self.name = 'Index_policy'
        if type(pre_indexs) is list:
            self.pre_indexs = pre_indexs
        else:
            raise ValueError("pre_indexs should be list")

        if type(post_indexs) is list:
            self.post_indexs = post_indexs
        else:
            raise ValueError("post_indexs should be list")

        if len(pre_indexs) != len(post_indexs):
            raise ValueError(" the length of pre and post index is not equal")
        else:
            self.index_len = len(pre_indexs)

    def checked_connection(self, new_connection: Projection):

        pre_level_groups = new_connection.pre.get_leveled_groups()
        post_level_groups = new_connection.post.get_leveled_groups()
        for ind in range(self.index_len):
            pre_ind = self.pre_indexs[ind]
            post_ind = self.post_indexs[ind]
            if not hasattr(pre_ind,'__iter__'):
                pre_group = pre_level_groups[2][pre_ind]
            else:
                pre_group = pre_level_groups[pre_ind[0]][pre_ind[1]]
            if not hasattr(post_ind,'__iter__'):
                post_group = post_level_groups[2][post_ind]
            else:
                post_group = post_level_groups[post_ind[0]][post_ind[1]]
            new_connection.add_connection(Projection(pre_group, post_group, level=1))

        return new_connection


#
# asb1 = Assembly()
# asb2 = Assembly()
#
#
# class Group1(Assembly):
#     def __init__(self, name=None):
#         super(Group1, self).__init__(name)
#
#         self.n1 = NeuronGroup(100 ,model='clif')
#         self.n2 = NeuronGroup(100, model='clif')
#         self.n1.add_type('in')
#         self.n2.add_type('out')
#
#
# with asb1:
#     g1 = NeuronGroup(100 ,model='clif')
#     g2 = NeuronGroup(100 ,model='clif')
#     g3 = Group1()
# asb1.g1.add_type('in')
#
#
# with asb2:
#     g4 = NeuronGroup(100 ,model='clif')
#     g5 = NeuronGroup(100 ,model='clif')
#     g6 = Group1()
#
# asb2.g4.add_type('out')
#
#
#
#
#
#
# p1 = IndexConnectPolicy(pre_indexs=[(1,1), (2,0)], post_indexs=[(2,1), (1,0)])
# c = p1.generate_connection(asb1, asb2)
# print(c)
#












# -*- coding: utf-8 -*-
"""
Created on 2020/8/5
@project: SPAIC
@filename: Assembly
@author: Hong Chaofei
@contact: hongchf@gmail.com
@description:
"""
import spaic

from collections import OrderedDict
from spaic.Network.BaseModule import BaseModule, VariableAgent
from abc import ABC, abstractmethod
from torch import nn
from typing import List
import inspect


# class ContextMetaClass(type):
#
#     def __call__(self, *args, **kwargs):
#         spaic.global_assembly_init_count += 1
#         rv = super(ContextMetaClass, self).__call__(*args, **kwargs)
#         spaic.global_assembly_init_count -= 1
#         return rv

class Assembly(BaseModule):
    r"""Base class for all network units.

        The Assembly represent an abstract network structure, it defines the basic network construct behavior and attributes.
        It can contain any neural network units like neurongroups, nodes, other assemblies and their connections.
        The classes like Node, NeuronGroup, and Network are special cases of Assembly.

        Methods:
            add_assembly: Add a new assembly to this assembly as its member.
            del_assembly: Delete an existed member assembly of this assembly.
            add_connection: Add the connection between two member assemblies of this assembly.
            del_connection: Delete an existed connection between member assemblies of this assembly.
            copy_assembly: Copy an existed assembly structure into this assembly.
            replace_assembly: Replace an existed member assembly with a new assembly.
            merge_assembly:   Add the member assemblies and connections of the target assembly, which are not already included in this assembly, to this assembly.
            select_assembly:   Select a list of member assemblies in this assembly, and form a new assembly that contains those selected assemblies.



        Attributes:
            _class_label(str): the label is a static variable to imply the class of this object
            _backend(Backend): the backend backend this assembly runs on
            _groups(OrderedDict): the container of member assemblies
            _connections(OrderedDict): the container of member connections
            _supers(list): the super assemblies that add this assembly as their member assemblies
            _input_connections(list): the connections that use this assembly as post-synaptic target
            _output_connections(list): the connections that use this assembly as pre-synaptic target
            num(int): the total number of neurons this assembly contains
            position(Tuple(int, int)): the top level positon of this assembly
            _var_names: The backend variable names this assembly and its members contains


    """

    _class_label = '<asb>'
    _is_terminal = False

    def __init__(self, name=None):
        '''Base class for  neural network units, and defines the basic attributes of a group object.

        It contains other Assemblies (including neuron_groups, networks and nodes) and their connections.
        This class provides a abstract representation of network topology.

        Args:
           name(str): name of the network assembly

        Attributes:
            _backend(Backend): the backend backend this assembly runs on
            _groups(OrderedDict): the container of member assemblies
            _connections(OrderedDict): the container of member connections
            _supers(list): the super assemblies that add this assembly as their member assemblies
            _input_connections(list): the connections that use this assembly as post-synaptic target
            _output_connections(list): the connections that use this assembly as pre-synaptic target
            num(int): the total number of neurons this assembly contains
            position(Tuple(int, int)): the top level positon of this assembly
            _var_names: The backend variable names this assembly and its members contains
        '''
        super(Assembly, self).__init__()
        context = spaic.global_assembly_context_list
        init_count = spaic.global_assembly_init_count

        self.set_name(name)
        self._backend: spaic.Backend = None
        self._groups: OrderedDict[str, Assembly] = OrderedDict()
        self._connections: OrderedDict[str, spaic.Connection] = OrderedDict()
        self._projections: OrderedDict[str, spaic.Projection] = OrderedDict()
        self._supers = list()
        self._input_connections = list()
        self._output_connections = list()
        self._input_modules = list()
        self._output_modules = list()
        self.num = 0
        self.position = None
        self.model_name = None
        # _var_names 和 _var_dict移动到了BaseModule 作为网络各模块的通用属性
        # self._var_names = []
        # self._var_dict = dict()
        self.context_enterpoint = 0
        self.type = []

    # front-end functions
    def add_type(self, type):
        self.type.append(type)

    def get_labeled_name(self, key: str):
        if isinstance(key, str):
            if '[updated]' in key:
                return self.id + ':' +'{'+key.replace('[updated]',"")+'}' + '[updated]'
            else:
                return self.id + ':' +'{'+key+'}'
        elif isinstance(key, VariableAgent):
            return key.var_name
        elif isinstance(key, list) or isinstance(key, tuple):
            keys = []
            for k in key:
                if isinstance(k, str):
                    if '[updated]' in k:
                        mk = self.id + ':' + '{' + k.replace('[updated]', "") + '}' + '[updated]'
                    else:
                        mk = self.id + ':' + '{' + k + '}'
                    keys.append(mk)
                elif isinstance(k, VariableAgent):
                    keys.append(k.var_name)
            return keys

    def add_assembly(self, name, assembly):
        """
        Add a new assembly to this assembly as its member.
        This is the basic method to build the network structure, neurongroups, nodes and other assemblies is added
        to this assembly by calling add_assembly function, explicitly or implicitly.

        Args:
            name (str): the attribute name of the added assembly
            assembly (Assembly): the assembly to be added

        Returns:
            None

        Examples:

            if you want to add a new assembly to in assembly initialization, you can explicitly use add_assembly:

            >>> def __init__(self, name=None):
            >>>    ...
            >>>    self.add_assembly(name='layer1', assembly=Assembly())

            or implicitly use add_assembly:

            >>> def __init__(self, name=None):
            >>>     ...
            >>>     self.layer1 = Assembly()

            you can also add new assemblies after the target assembly has been created:

            >>> TestAsb = Assembly()
            >>> TestAsb.add_assembly(name='layer1', assembly=Assembly())
        """

        if self._backend: self._backend.builded = False

        assert assembly not in self._groups.values(), "assembly %s is already in the assembly %s"%(name, self.name)
        assert name not in self._groups, "assembly with name: %s have the same name with assembly already in the assembly %s"%(name, self.name)
        self.__setattr__(name, assembly)

    def del_assembly(self, assembly=None, name=None):
        """
        Delete an existed member assembly of this assembly.

        User can delete by assembly object or delete by assembly name.

        Args:
            assembly(Assembly): the member assembly to be deleted
            name(str): the member name of the assembly to be deleted
        Returns:
            None

        Examples:
            User can delete member assembly by object:

            >>> TestAsb = Assembly() # assuming contains neurongroups and network structure
            >>> TestAsb.del_assembly(assembly=TestAsb.layer1)

            User can also delete member assembly by the name:

            >>> TestAsb = Assembly() # assuming contains neurongroups and network structure
            >>> TestAsb.del_assembly(name='layer1')
        """
        if self._backend: self._backend.builded = False

        if assembly is not None:
            deleted_assembly = False
            for gkey, value in self._groups.items():
                if value is assembly:
                    del self._groups[gkey]
                    del self.__dict__[gkey]
                    deleted_assembly = True
            assert deleted_assembly, " try to delete an assembly that is not in the group"

            for ckey in self._connections.keys():
                if self._connections[ckey].assembly_linked(assembly):
                    del self._connections[ckey]

        elif name is not None:
            assert name in self._groups, " try to delete an assembly that is not in the group"
            assembly = self._groups[name]
            del self._groups[name]
            del self.__dict__[name]
            for ckey in self._connections.keys():
                if self._connections[ckey].assembly_linked(assembly):
                    del self._connections[ckey]

    def add_connection(self, name, connection):
        """
        Add the connection between two member assemblies of this assembly.

        Args:
            name(str): name of this connection
            connection(Connection): the new connection to be added to the assembly

        Returns:
            None

        Examples:

            >>> TestAsb = Assembly() # assuming contains neurongroups and network structure
            >>> TestAsb.add_connection(name='con1', connection=Connection(self.layer1, self.layer2, link_type='full'))

        """
        if self._backend: self._backend.builded = False
        # assert connection.pre in self.get_groups(), 'pre %s is not in the group' % connection.pre.name
        # assert connection.post in self.get_groups(), 'post %s is not in the group' % connection.post.name
        if name in self._connections:
            if connection is self._connections[name]:
                raise ValueError(" connection is already in the assembly's connection list")
            else:
                raise ValueError("duplicated name for the connection")
        else:
            # self._connections[name] = connection
            self.__setattr__(name, connection)

    def del_connection(self, connection=None, name=None):
        """
        Delete an existed connection between member assemblies of this assembly.

        User can delete the connection by connection object or by connection name.

        Args:
            connection(Connection): the connection object to be deleted
            name(str): the name of connection to be deleted

        Returns:
            None

        Examples:
            Delete by object:

            >>> TestAsb = Assembly() # assuming it contains neurongroups and network structure
            >>> TestAsb.del_connection(connection=TestAsb.con1)

            Delete by name:

            >>> TestAsb = Assembly()
            >>> TestAsb.del_connection(name='con1')
        """
        if self._backend: self._backend.builded = False
        if connection is not None:
            deleted_connection = False
            for ckey, value in self._connections.items():
                if value is connection:
                    dict.__getitem__()
                    del self._connections[ckey]
                    del self.__dict__[ckey]
                    deleted_connection = True 
            assert deleted_connection,  " try to delete an connection that is not in the group"
        elif name is not None:
            assert name in self._connections, " try to delete an connection that is not in the group"
            del self._connections[name]
            del self.__dict__[name]

    def add_projection(self, name, projection):
        """
        Add the projection between two member assemblies of this assembly.

        Args:
            name(str): name of this projection
            projection(Projection): the new projection to be added to the assembly

        Returns:
            None

        Examples:

            >>> TestAsb = Assembly() # assuming contains neurongroups and network structure
            >>> TestAsb.add_projection(name='prj1', projection=Projection(self.layer1, self.layer2, link_type='full'))

        """
        if self._backend: self._backend.builded = False
        assert projection.pre in self._groups.values(), 'pre %s is not in the group' % projection.pre.name
        assert projection.post in self._groups.values(), 'post %s is not in the group' % projection.post.name
        if name in self._projections:
            if projection is self._projections[name]:
                raise ValueError(" projection is already in the assembly's projection list")
            else:
                raise ValueError("duplicated name for the projection")
        else:
            self.__setattr__(name, projection)

    def copy_assembly(self, name, assembly):
        """
        Copy an existed assembly structure into this assembly.

        A new assembly is initialized which copy the structure(type and connection of the assembly members) of the original assembly,
        and the new assembly is added to this assembly.

        Args:
            name(str): the name of the new copy assembly
            assembly(Assembly): the assembly object to be copied

        Returns:
            None

        Examples:

            >>> Asb1 = Assembly() # assuming it contains neurongroups and network structure
            >>> Asb2 = Assembly() # assuming it contains neurongroups and network structure
            >>> Asb1.copy_assembly(name='layer2', assembly=Asb2)
        """
        if self._backend: self._backend.builded = False
        rv = assembly.structure_copy(name)
        self.__setattr__(name, rv)

    def replace_assembly(self, old_assembly, new_assembly):
        """
        Replace an existed member assembly with a new assembly.

        Delete the existed old member assembly, add the new assembly to this assembly,
        and redirect related connections from the old assembly to the new assembly.

        Args:
            old_assembly(Assembly): the old member assembly to be replaced
            new_assembly(Assembly): the new assembly

        Returns:
            None

        Examples:

            >>> templateAsb = Assembly() # assuming templateAsb contains a member assembly called asb1
            >>> asb2 = Assembly() # assuming it contains neurongroups and network structure
            >>> templateAsb.replace_assembly(templateAsb.asb1, asb2)
        """
        if self._backend: self._backend.builded = False
        replaced_assembly = False
        for gkey in self._groups.keys():
            if self._groups[gkey] is old_assembly:
                self._groups[gkey] = new_assembly
                self.__dict__[gkey] = new_assembly
                replaced_assembly = True
        assert replaced_assembly, " try to repalce an assembly that is not in the group"

        for con in self._connections.values():
            if con.assembly_linked(old_assembly):
                con.replace_assembly(old_assembly, new_assembly)

    def merge_assembly(self, assembly):
        """
        Add the member assemblies and connections of the target assembly, which are not already included in this assembly, to this assembly.


        Args:
            assembly(Assembly): the target assembly, from which this assembly will copy member assemblies and connections.

        Returns:
            None

        Examples:

            >>> target_asb = Assembly()  # assuming it contains neurongroups and network structure
            >>> test_asb = Assembly() # assuming it contains neurongroups and network structure
            >>> test_asb.merge_assembly(target_asb)
        """
        if self._backend: self._backend.builded = False

        for key, value in assembly._groups.item():
            if key in self._groups:
                # for the different sub_assembly with different name:
                if value is not self._groups[key]:
                    # set a duplicated name with suffix (1),(2),(3),...
                    for num in range(1, 1000):
                        tmp_key = key + '(' + str(num) +')'
                        if tmp_key not in self._groups:
                            self._groups[tmp_key] = value
                            break
            else:
                # for the sub_assembly that is not in self
                self._groups[key] = value

        for key, value in assembly._connections.item():
            if key in self._connections:
                if value is not self._connections[key]:
                    # set a duplicated name with suffix (1),(2),(3),...
                    for num in range(1, 1000):
                        tmp_key = key + '(' + str(num) +')'
                        if tmp_key not in self._connections:
                            self._connections[tmp_key] = value
                            break
            else:
                self._connections[key] = value

    def select_assembly(self, assemblies, name=None, with_connection=True):
        """
        Select a list of member assemblies in this assembly, and form a new assembly that contains those selected assemblies
        and their connections(if with_connection is True).

        Args:
            assemblies(List[Assembly]): list of assemblies (or member assembly names) to be selected to form a new assembly
            name: the name of the new assembly

        Returns:
            a new assembly that contains the selected assemblies and their connections (if with_coonection is True)

        Examples:

            >>> testAsb = Assembly() # assuming it contains member assemblies named asb1, asb2, asb3...
            >>> newAsb1 = testAsb.select_assembly(['asb1', 'asb2'], 'newAsb') # using names
            >>> newAsb2 = testAsb.select_assembly([testAsb.asb2, testAsb.asb3], 'newAsb') # using assembly objects
        """
        if self._backend: self._backend.builded = False
        new_asb = Assembly(name)
        for asb in assemblies:
            if isinstance(asb, str):
                asb_name= asb
                if asb_name in self._groups:
                    asb = self._groups[asb_name]
                    new_asb.add_assembly(asb_name, asb)
                    if with_connection:
                        for key, con in self._connections.items():
                            if con.assembly_linked(asb):
                                new_asb.add_connection(key, con)
                else:
                    ValueError("No assembly name in the groups")
            else:
                assert isinstance(asb, Assembly), "selected object that is not Assembly"
                for gkey, value in self._groups:
                    if value is asb:
                        new_asb.add_assembly(asb.name, asb)
                        if with_connection:
                            for key, con in self._connections.items():
                                if con.assembly_linked(asb):
                                    new_asb.add_connection(key, con)
                else:
                    ValueError("No assembly in the groups")

        return new_asb

    def assembly_hide(self):
        """
        Prohibit this assembly from building and display, but keep this assembly for later use.

        The set this assembly and its member assemblies with the flag hided = True.

        Returns:
            None

        Examples:
            >>> TestAsb = Assembly()
            >>> TestAsb.assembly_hide()
        """
        if self._backend: self._backend.builded = False
        self.enabled = False
        for key, value in self._groups.items():
            value.assembly_hide()
        for key, value in self._connections.items():
            value.enabled = False

    def assembly_show(self):
        """
        Make the hided assembly to normal assembly.

        Returns:
            None

        Examples:
        >>> TestAsb = Assembly() # assuming hided
        >>> TestAsb.assembly_show()

        """
        if self._backend: self._backend.builded = False
        self.enabled = True
        for key, value in self._groups.items():
            value.assembly_show()
        for key, value in self._connections.items():
            value.enabled = True

    def get_groups(self, recursive=True):
        """
        Get all member neurongroups and neurongroups in member assemblies in a list.
        Args:
            recursive(bool): flag to decide if members of the member assemblies should be returned.

        Returns:
            list of all member groups

        """
        if self._groups and recursive:
            all_groups = []
            for g in self._groups.values():
                all_groups.extend(g.get_groups(recursive))
            return all_groups
        elif self._groups and not recursive:
            return list(self._groups.values())
        elif self._class_label == '<asb>' or self._class_label == '<net>':
            return []
        else:
            return [self]

    def get_leveled_groups(self):
        """
        Get list of all sup groups in leveled order, such as [ [self], [subgroups], [subgroup of subgroups], ...]
        Returns:

        """
        if self._is_terminal:
            return [[self]]
        else:
            leveled_groups = [[self]]
            for k, g in self._groups.items():
                new_level_groups = g.get_leveled_groups()
                for level, groups in enumerate(new_level_groups):
                    level += 1
                    if level < len(leveled_groups):
                        leveled_groups[level].extend(groups)
                    else:
                        leveled_groups.append(groups)
        return leveled_groups

    def get_assemblies(self, recursive=True, include_empty=False):
        """
        Get all the member assemblies and assemblies in member assemblies.
        Args:
            recursive(bool): flag to decide if members of the member assemblies should be returned.

        Returns:
            list of all member assemblies
        """

        if type(recursive) is int:
            # use recursive as a level label
            recursive -= 1

        if self._groups and recursive:
            all_assemblies = {self,}
            for g in self._groups.values():
                all_assemblies.update(g.get_assemblies(recursive, include_empty))
            return all_assemblies
        elif self._groups and not recursive:
            return {self}
        elif not self._groups and include_empty:
            return {self}
        else:
            return []

    def get_assembly_key(self, assembly):
        """
        Get the key of the target assembly if it is a member of this assembly
        Args:
            assembly: the target assembly

        Returns:
            the key of target assembly
        """
        for gkey in self._groups.keys():
            if self._groups[gkey] is assembly:
                return gkey
        return False

    def get_super_assemblies(self, assembly):
        """
        Get all the super assembly of the target assembly if it is a member or member's member of self assembly
        Args:
            assembly: the target assembly

        Returns:
            list of super assemblies or []
        """
        if len(self._groups) == 0:
            return []
        elif assembly in self._groups.values():
            return [self]
        else:
            for g in self._groups.values():
                sup_list = g.get_super_assemblies(assembly)
                if sup_list:
                    return_list = [self]
                    return_list.extend(sup_list)
                    return return_list
            return []

    def __contains__(self, item):
        if isinstance(item, Assembly):
            if item is self:
                return True
            elif item in self._groups.values():
                return True
            else:
                for g in self._groups.values():
                    if item in g:
                        return True
                return False
        else:
            return item in self._connections.values()

    def get_connections(self, recursive=True):
        """
            Get the Connections in this assembly
        Args:
            recursive(bool): flag to decide if member connections of the member assemblies should be returned.

        Returns:
            List of Connections
        """
        if not recursive:
            return list(self._connections.values())
        else:
            all_assmblies = self.get_assemblies(recursive=2)
            connections = list()
            connections = self.update_connection(connections, self._connections)
            # connections.update(self._connections.values())
            for asb in all_assmblies:
                if asb is self:
                    # connections.update(asb.get_connections(recursive=False))
                    connections = self.update_connection(connections, asb.get_connections(recursive=False))
                else:
                    # connections.update(asb.get_connections(recursive=True))
                    connections = self.update_connection(connections, asb.get_connections(recursive=True))
            for proj in self._projections.values():
                # connections.update(proj.get_connections(recursive=True))
                connections = self.update_connection(connections, proj.get_connections(recursive=True))
            return connections
    def update_connection(self, container, connections):
        assert isinstance(container, list)
        if isinstance(connections, OrderedDict):
            for con in connections.values():
                if con not in container:
                    container.append(con)
        elif isinstance(connections, list):
            for con in connections:
                if con not in container:
                    container.append(con)
        else:
            raise ValueError("connections type not right")
        return container


    def get_var_names(self):
        """
        Get a list of variable names the assembly member contains.

        """
        return self._var_names

    def get_str(self, level):
        """
        Get a string description of the strcuture of this assembly
        Args:
            level: the deepth of this assembly relative to the top network

        Returns:
            String representations
        """

        level_space = "" + '-'*level
        repr_str = level_space + "|name:{}, type:{}, ".format(self.name, type(self).__name__)
        repr_str += "total_neuron_num:{}\n ".format(self.num)
        level += 1
        for g in self._groups.values():
            repr_str += g.get_str(level)
        for c in self._connections.values():
            repr_str += c.get_str(level)
        for p in self._projections.values():
            repr_str += p.get_str(level)
        return repr_str

    # back-end functions
    def build(self, backend=None, strategy=0):
        """
        Build the front-end network structure into a back-end computation graph.

        Args:
            backend(Backend): the backend backend to be builded into

        Returns:
            None

        """
        self._backend = backend
        print("builder for assembly has been called")
        # for asb in self.get_groups():
        #     asb.set_id()
        # for con in self.get_connections():
        #     con.set_id()

        for key, value in self._connections.items():
            value.build(backend)
        for key, value in self._groups.items():
            value.build(backend)

    def build_projections(self, backend):
        for proj in self._projections.values():
            proj.build(backend)
        for group in self._groups.values():
            group.build_projections(backend)

    def set_id(self):
        """
        Get the ID of this assembly
        """

        if self.id is not None:
            return self.id

        if len(self._supers) == 0:
            self.id = self.name + self.__class__._class_label
        else:
            super_ids = []
            for super in self._supers:
                if super.id is not None:
                    super_ids.append(super.id)
                else:
                    super_ids.append(super.set_id())

            self.id = self.name + self.__class__._class_label
            if len(super_ids) == 1:
                self.id = super_ids[0] + '_' + self.id
            else:
                pre_id = '/'
                for prefix in super_ids:
                    pre_id += prefix + ','
                pre_id += '/'
                self.id = pre_id + '_' + self.id
        return self.id

    def register_connection(self, connection_obj, presynaptic):
        '''
        Register input or output connection of this assembly
        Args:
            connection_obj (Connection.Connection): the connection
            presynaptic (bool) : if this assembly is presynaptic neuron
        Returns:
            None

        '''
        # ###对复杂结构的assembly也进行注册
        # for key, i in self._groups.items():
        #     if presynaptic:
        #         if connection_obj not in i._output_connections:
        #             i._output_connections.append(connection_obj)
        #     else:
        #         if connection_obj not in i._input_connections:
        #             i._input_connections.append(connection_obj)
        # # connection_obj.post_groups
        if presynaptic:
            if connection_obj not in self._output_connections:
                self._output_connections.append(connection_obj)
        else:
            if connection_obj not in self._input_connections:
                self._input_connections.append(connection_obj)

    def register_module(self, module_obj, pre):
        if pre:
            if module_obj not in self._output_modules:
                self._output_modules.append(module_obj)
        else:
            if module_obj not in self._input_modules:
                self._input_modules.append(module_obj)


    def structure_copy(self, name=None):
        """
        Copy the structure of this assembly with new members
        Args:
            name: name of the new Assembly

        Returns:
            the new assembly
        """
        # define a new object but remain the structure
        from copy import deepcopy
        rv = self.__class__.__new__(self.__class__)
        rv.__init__(name)
        tmp_dict = deepcopy(self.__dict__)
        del tmp_dict['name']
        del tmp_dict['_supers']
        del tmp_dict['_input_connections']
        del tmp_dict['_output_connections']
        rv.__dict__.update(tmp_dict)
        return rv

    def add_super(self, assembly):
        '''
        Tell this assemlby  the target assembly is it's super assembly
        Args:
            assembly: the target super assembly
        '''
        assert isinstance(assembly, Assembly), "the super is not Assembly"
        self._supers.append(assembly)

    def del_super(self, assembly):
        """
        Delete certain assembly from super assemblies of this assembly
        Args:
            assembly: the target super assembly

        """
        assert  assembly in self._supers, "the assembly is not in supers"
        self._supers.remove(assembly)


    # def __getattr__(self, name):

    def train(self, mode=True):
        self.training = mode
        for g in self._groups.values():
            g.train(mode)
        for p in self._projections.values():
            p.train(mode)
        for c in self._connections.values():
            c.training = mode

    def eval(self):
        self.train(False)


    def __enter__(self):
        import __main__
        spaic.global_assembly_context_list.append(self)
        main_vars = vars(__main__)
        NoInMain = True
        for key in main_vars:
            value = main_vars[key]
            if value is self:
               NoInMain = False
               self.set_name(key)
               break

        if NoInMain:
            raise ValueError("can only construct network using with at __main__")
        else:
            # record the variable number before enter the context
            self.context_enterpoint = main_vars.__len__() -1

    def __exit__(self, exc_type, exc_val, exc_tb):
        import __main__
        main_vars = vars(__main__)
        endpoint_num = main_vars.__len__() -1

        # depending on the feature that python >=3.7 , dict is insertion ordered, so we can get the subunits by order
        for ind, key in enumerate(main_vars):
            if ind > self.context_enterpoint and ind <= spaic.global_assembly_context_omit_start:
                self.__setattr__(key, main_vars[key])
            elif ind > self.context_enterpoint and ind > spaic.global_assembly_context_omit_end:
                self.__setattr__(key, main_vars[key])

        spaic.global_assembly_context_list.pop()
        spaic.global_assembly_context_omit_start = self.context_enterpoint
        spaic.global_assembly_context_omit_end = endpoint_num
        # keys = list(globals().keys())
        # print(keys)

    def __setattr__(self, name, value):
        from spaic.Network.Topology import Connection
        from spaic.Network.Topology import Projection
        super(Assembly, self).__setattr__(name, value)
        if (self.__class__ is spaic.NeuronGroup) or (issubclass(self.__class__, spaic.Node)):
            # If class is NeuronGroup or the subclass of Node, do not add other object to it.
            return

        if isinstance(value, Assembly):
            if self._backend: self._backend.builded = False
            value.set_name(name)
            self._groups[name] = value
            # self.num += value.num
            value.add_super(self)
        elif isinstance(value, Connection):
            if self._backend: self._backend.builded = False
            self._connections[name] = value
            value.set_name(name)
            value.add_super(self)
        elif isinstance(value, Projection):
            # if it is not Connection but belongs to projection (pure projection)
            if self._backend: self._backend.builded = False
            self._projections[name] = value
            value.set_name(name)
            value.add_super(self)

    def __delattr__(self, name):
        super(Assembly, self).__delattr__(name)
        if name in self._groups:
            if self._backend: self._backend.builded = False
            self._groups[name].del_super(self)
            del self._groups[name]
        elif name in self._connections:
            if self._backend: self._backend.builded = False
            self._connections[name].del_super(self)
            del self._connections[name]

    def __repr__(self):

        repr_str = self.get_str(0)
        return repr_str



# Copyright 2014 Jason Heeris, jason.heeris@gmail.com
#
# This file is part of the gammatone toolkit, and is licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING
"""
Plotting utilities related to gammatone analysis, primarily for use with
``matplotlib``.
"""
from __future__ import division
import argparse
import os.path

import matplotlib.pyplot
import matplotlib.ticker
import numpy as np
import scipy.constants
import scipy.io.wavfile

from .filters import erb_point
from .gtgram import gtgram
from .fftweight import fft_gtgram


class ERBFormatter(matplotlib.ticker.EngFormatter):
    """
    Axis formatter for gammatone filterbank analysis. This formatter calculates
    the ERB spaced frequencies used for analysis, and renders them similarly to
    the engineering axis formatter.

    The scale is changed so that `[0, 1]` corresponds to ERB spaced frequencies
    from ``high_freq`` to ``low_freq`` (note the reversal). It should be used
    with ``imshow`` where the ``extent`` argument is ``[a, b, 1, 0]`` (again,
    note the inversion).
    """

    def __init__(self, low_freq, high_freq, *args, **kwargs):
        """
        Creates a new :class ERBFormatter: for use with ``matplotlib`` plots.
        Note that this class does not supply the ``units`` or ``places``
        arguments; typically these would be ``'Hz'`` and ``0``.

        :param low_freq: the low end of the gammatone filterbank frequency range
        :param high_freq: the high end of the gammatone filterbank frequency
          range
        """
        self.low_freq = low_freq
        self.high_freq = high_freq
        super().__init__(*args, **kwargs)

    def _erb_axis_scale(self, fraction):
        return erb_point(self.low_freq, self.high_freq, fraction)

    def __call__(self, val, pos=None):
        newval = self._erb_axis_scale(val)
        return super().__call__(newval, pos)


def gtgram_plot(
        gtgram_function,
        axes, x, fs,
        window_time, hop_time, channels, f_min,
        imshow_args=None
        ):
    """
    Plots a spectrogram-like time frequency magnitude array based on gammatone
    subband filters.

    :param gtgram_function: A function with signature::

        fft_gtgram(
            wave,
            fs,
            window_time, hop_time,
            channels,
            f_min)

    See :func:`gammatone.gtgram.gtgram` for details of the paramters.
    """
    # Set a nice formatter for the y-axis
    formatter = ERBFormatter(f_min, fs/2, unit='Hz', places=0)
    axes.yaxis.set_major_formatter(formatter)

    # Figure out time axis scaling
    duration = len(x) / fs

    # Calculate 1:1 aspect ratio
    aspect_ratio = duration/scipy.constants.golden

    gtg = gtgram_function(x, fs, window_time, hop_time, channels, f_min)
    Z = np.flipud(20 * np.log10(gtg))

    img = axes.imshow(Z, extent=[0, duration, 1, 0], aspect=aspect_ratio)


# Entry point for CLI script

HELP_TEXT = """\
Plots the gammatone filterbank analysis of a WAV file.

If the file contains more than one channel, all channels are averaged before
performing analysis.
"""


def render_audio_from_file(path, duration, function):
    """
    Renders the given ``duration`` of audio from the audio file at ``path``
    using the gammatone spectrogram function ``function``.
    """
    samplerate, data = scipy.io.wavfile.read(path)

    # Average the stereo signal
    if duration:
        nframes = duration * samplerate
        data = data[0 : nframes, :]

    signal = data.mean(1)

    # Default gammatone-based spectrogram parameters
    twin = 0.08
    thop = twin / 2
    channels = 1024
    fmin = 20

    # Set up the plot
    fig = matplotlib.pyplot.figure()
    axes = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    gtgram_plot(
        function,
        axes,
        signal,
        samplerate,
        twin, thop, channels, fmin)

    axes.set_title(os.path.basename(path))
    axes.set_xlabel("Time (s)")
    axes.set_ylabel("Frequency")

    matplotlib.pyplot.show()


def main():
    """
    Entry point for CLI application to plot gammatonegrams of sound files.
    """
    parser = argparse.ArgumentParser(description=HELP_TEXT)

    parser.add_argument(
        'sound_file',
        help="The sound file to graph. See the help text for supported formats.")

    parser.add_argument(
        '-d', '--duration', type=int,
        help="The time in seconds from the start of the audio to use for the "
             "graph (default is to use the whole file)."
        )

    parser.add_argument(
        '-a', '--accurate', action='store_const', dest='function',
        const=gtgram.gtgram, default=fft_gtgram,
        help="Use the full filterbank approach instead of the weighted FFT "
             "approximation. This is much slower, and uses a lot of memory, but"
             " is more accurate."
        )

    args = parser.parse_args()

    return render_audio_from_file(args.sound_file, args.duration, args.function)



# coding: utf-8
from setuptools import setup

# Metadata goes in setup.cfg. There are here for GitHub's dependency graph
setup(
    name='SPAIC',
    install_requires=[
        'torch>=1.5.0',
        'matplotlib',
        'numpy',
        'tqdm',
        'scipy',
        'pyyaml',
        'h5py',
    ],
)



# -*- coding: utf-8 -*-
"""
Created on 2021/4/12
@project: SPAIC
@filename: Module
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
wrap around deep learning module such as a cnn network lstm cell
"""
import torch

import spaic
from spaic.Network.Assembly import Assembly
from spaic.Network.BaseModule import Op

class Module(Assembly):
    _class_label = '<mod>'

    def __init__(self, module=None, name=None, input_targets=[], input_var_names=['O[updated]'], output_targets=None, output_var_names=['Isyn'], module_backend='pytorch'):
        super(Module, self).__init__(name)
        self.module: torch.nn.Module = module
        if isinstance(input_targets, list):
            self.input_targets = input_targets
        else:
            self.input_targets = [input_targets]
        if isinstance(input_var_names, list):
            self.input_var_names = input_var_names
        else:
            self.input_var_names = [input_var_names]

        if isinstance(output_targets, list):
            self.output_targets = output_targets
        else:
            self.output_targets = [output_targets]
        if isinstance(output_var_names, list):
            self.output_var_names = output_var_names
        else:
            self.output_var_names = [output_var_names]

        for in_targ in self.input_targets:
            in_targ.register_module(self, True)
        for out_targ in self.output_targets:
            out_targ.register_module(self, False)

        self.module_backend = module_backend


    def standalone_run(self, *args):
        return self.module(*args)

    def init_variable(self, var_names=None, var_shapes=None, var_value_dict=None):
        if var_names is None:
            self._var_names = []
        elif hasattr(var_names, '__iter__'):
            self._var_names = var_names
        else:
            self._var_names = [var_names]

        if var_shapes is None:
            self.var_shapes = []
        elif isinstance(var_shapes, list):
            self._var_shapes = var_shapes
        else:
            self._var_shapes = [var_shapes for _ in range(len(self._var_names))]
        self._var_values = []

        if var_value_dict is None:
            var_value_dict = []
        elif hasattr(var_value_dict, '__iter__'):
            var_value_dict = var_value_dict
        else:
            var_value_dict = [var_value_dict]
        for var_name in self._var_names:
            if var_name in var_value_dict:
                self._var_values.append(var_value_dict[var_name])
            else:
                self._var_values.append(0.0)

    def build(self, backend):

        # Add module owned variables to backend
        self._backend = backend
        self.init_variable()
        var_len = len(self._var_names)
        for ii in range(var_len):
            key = self.id + ":" + "{" + self._var_names[ii] + "}"
            shape = (1, *self._var_shapes[ii])
            self.variable_to_backend(key, shape, self._var_values[ii])

        # add standalone operation
        output_var_name = self.output_targets[0].id + ":" + "{" + self.output_var_names[0] + "}"
        self._var_names.append(output_var_name)
        input_var_names = []
        for input_target, input_name in zip(self.input_targets, self.input_var_names):
            input_var_name = input_target.id + ":" + "{" + input_name + "}"
            self._var_names.append(input_var_name)
            input_var_names.append(input_var_name)

        backend.register_standalone(Op(output_var_name, self.standalone_run, input_var_names, owner=self))
        self.module.to(backend.device0)



    @property
    def parameters(self):

        return self.module.parameters()  # .state_dict()

    @property
    def state_dict(self):

        return self.module.state_dict()

    def load_state_dict(self, state):

        return self.module.load_state_dict(state)

    def train(self, mode=True):
        self.module.train(mode)



# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: surrogate.py
@time:2022/8/12 17:14
@description:
"""
import math
import torch
from torch import nn
from torch.nn import functional as F


def heaviside(input):
    return (input >= 0.).to(input.dtype)


class SurrogateFunctionBase(nn.Module):
    """
    Surrogate Function 的基类
    :param alpha: 为一些能够调控函数形状的代理函数提供参数.
    :param requires_grad: 参数 ``alpha`` 是否需要计算梯度, 默认为 ``False``
    """
    def __init__(self, alpha, requires_grad=True):
        super().__init__()
        self.alpha = nn.Parameter(
            torch.tensor(alpha, dtype=torch.float),
            requires_grad=requires_grad)

    @staticmethod
    def firing_func(input, alpha):
        """
        :param input: 膜电位的输入
        :param alpha: 控制代理梯度形状的变量, 可以为 ``NoneType``
        :return: 激发之后的spike, 取值为 ``[0, 1]``
        """
        raise NotImplementedError

    def forward(self, input):
        """
        :param input: 膜电位输入
        :return: 激发之后的spike
        """
        return self.firing_func(input, self.alpha)


'''
    sigmoid surrogate function.
'''

class sigmoid(torch.autograd.Function):
    """
    使用 sigmoid 作为代理梯度函数
    对应的原函数为:
    .. math::
            g(input) = \\mathrm{sigmoid}(\\alpha input) = \\frac{1}{1+e^{-\\alpha input}}
    反向传播的函数为:
    .. math::
            g'(input) = \\alpha * (1 - \\mathrm{sigmoid} (\\alpha input)) \\mathrm{sigmoid} (\\alpha input)
    """
    @staticmethod
    def forward(ctx, input, alpha):
        if input.requires_grad:
            ctx.save_for_backward(input)
            ctx.alpha = alpha
        return heaviside(input)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        if ctx.needs_input_grad[0]:
            s_input = torch.sigmoid(ctx.alpha * ctx.saved_tensors[0])
            grad_input = grad_output * s_input * (1 - s_input) * ctx.alpha
        return grad_input, None


class SigmoidGrad(SurrogateFunctionBase):
    def __init__(self, alpha=1., requires_grad=False):
        super().__init__(alpha, requires_grad)

    @staticmethod
    def firing_func(input, alpha):
        return sigmoid.apply(input, alpha)


'''
    atan surrogate function.
'''


class atan(torch.autograd.Function):
    """
    使用 Atan 作为代理梯度函数
    对应的原函数为:
    .. math::
            g(input) = \\frac{1}{\\pi} \\arctan(\\frac{\\pi}{2}\\alpha input) + \\frac{1}{2}
    反向传播的函数为:
    .. math::
            g'(input) = \\frac{\\alpha}{2(1 + (\\frac{\\pi}{2}\\alpha input)^2)}
    """
    @staticmethod
    def forward(ctx, input, alpha):
        ctx.save_for_backward(input, alpha)
        return input.gt(0.).type_as(input)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        grad_alpha = None

        shared_c = grad_output / \
                   (1 + (ctx.saved_tensors[1] * math.pi /
                         2 * ctx.saved_tensors[0]).square())
        if ctx.needs_input_grad[0]:
            grad_input = ctx.saved_tensors[1] / 2 * shared_c
        if ctx.needs_input_grad[1]:
            grad_alpha = (ctx.saved_tensors[0] / 2 * shared_c).sum()

        return grad_input, grad_alpha


class AtanGrad(SurrogateFunctionBase):
    def __init__(self, alpha=2., requires_grad=True):
        super().__init__(alpha, requires_grad)

    @staticmethod
    def firing_func(input, alpha):
        return atan.apply(input, alpha)


'''
    gate surrogate fucntion. 
'''


class gate(torch.autograd.Function):
    """
    使用 gate 作为代理梯度函数
    对应的原函数为:
    .. math::
            g(input) = \\mathrm{NonzeroSign}(input) \\log (|\\alpha input| + 1)
    反向传播的函数为:
    .. math::
            g'(input) = \\frac{\\alpha}{1 + |\\alpha input|} = \\frac{1}{\\frac{1}{\\alpha} + |input|}
    """
    @staticmethod
    def forward(ctx, input, alpha):
        if input.requires_grad:
            grad_input = torch.where(input.abs() < 1. / alpha, torch.ones_like(input), torch.zeros_like(input))
            ctx.save_for_backward(grad_input)
        return input.gt(0).type_as(input)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        if ctx.needs_input_grad[0]:
            grad_input = grad_output * ctx.saved_tensors[0]
        return grad_input, None


class GateGrad(SurrogateFunctionBase):
    def __init__(self, alpha=2., requires_grad=False):
        super().__init__(alpha, requires_grad)

    @staticmethod
    def firing_func(input, alpha):
        return gate.apply(input, alpha)


'''
    quadratic_gate surrogate function.
'''


class quadratic_gate(torch.autograd.Function):
    """
    使用 quadratic_gate 作为代理梯度函数
    对应的原函数为:
    .. math::
        g(input) =
        \\begin{cases}
        0, & input < -\\frac{1}{\\alpha} \\\\
        -\\frac{1}{2}\\alpha^2|input|input + \\alpha x + \\frac{1}{2}, & |input| \\leq \\frac{1}{\\alpha}  \\\\
        1, & input > \\frac{1}{\\alpha} \\\\
        \\end{cases}
    反向传播的函数为:
    .. math::
        g'(input) =
        \\begin{cases}
        0, & |input| > \\frac{1}{\\alpha} \\\\
        -\\alpha^2|input|+\\alpha, & |input| \\leq \\frac{1}{\\alpha}
        \\end{cases}
    """
    @staticmethod
    def forward(ctx, input, alpha):
        if input.requires_grad:
            mask_zero = (input.abs() > 1 / alpha)
            grad_input = -alpha * alpha * input.abs() + alpha
            grad_input.masked_fill_(mask_zero, 0)
            ctx.save_for_backward(grad_input)
        return input.gt(0.).type_as(input)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        if ctx.needs_input_grad[0]:
            grad_input = grad_output * ctx.saved_tensors[0]
        return grad_input, None


class QGateGrad(SurrogateFunctionBase):
    def __init__(self, alpha=2., requires_grad=False):
        super().__init__(alpha, requires_grad)

    @staticmethod
    def firing_func(input, alpha):
        return quadratic_gate.apply(input, alpha)


class relu_like(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, alpha):
        if input.requires_grad:
            ctx.save_for_backward(input, alpha)
        return heaviside(input)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input, grad_alpha = None, None
        input, alpha = ctx.saved_tensors
        if ctx.needs_input_grad[0]:
            grad_input = grad_output * input.gt(0.).type_as(input) * alpha
        if ctx.needs_input_grad[1]:
            grad_alpha = (grad_output * F.relu(input)).sum()
        return grad_input, grad_alpha


class ReLUGrad(SurrogateFunctionBase):
    """
    使用ReLU作为代替梯度函数, 主要用为相同结构的ANN的测试
    """
    def __init__(self, alpha=2., requires_grad=False):
        super().__init__(alpha, requires_grad)

    @staticmethod
    def firing_func(input, alpha):
        return relu_like.apply(input, alpha)


'''
    Straight-Through (ST) Estimator
'''


class straight_through_estimator(torch.autograd.Function):
    """
    使用直通估计器作为代理梯度函数
    http://arxiv.org/abs/1308.3432
    """
    @staticmethod
    def forward(ctx, input):
        output = heaviside(input)
        ctx.save_for_backward(output)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        if ctx.needs_input_grad[0]:
            grad_input = grad_output
        return grad_input


# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: IO
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
"""
from abc import abstractmethod
from spaic.Neuron.Node import Action, Encoder, Reward
import numpy as np
from tqdm import tqdm
import torch
from collections import namedtuple, deque
import random

class Pipline:
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def update_step(self, batch, **kwargs):
        """
        Perform a pass of the network given the input batch.

        :param batch: The current batch. This could be anything as long as the subclass
            agrees upon the format in some way.
        :return: Any output that is need for recording purposes.
        """
        raise NotImplementedError("You need to provide a step_ method.")

class RLPipeline(Pipline):
    def __init__(self, network, environment, time=None, **kwargs):
        self.network = network
        self._backend = self.network._backend
        self.sim_name = self._backend.backend_name
        self.device = self._backend.device
        self.time = time
        self.time_step = int(self.time / self._backend.dt)
        self.actuator = None
        self.encoder = None
        self.rewarder = None
        # Get actuator and encoder
        for group in self.network.get_groups():
            if isinstance(group, Action):
                self.actuator = group
            if isinstance(group, Encoder):
                self.encoder = group
            if isinstance(group, Reward):
                self.rewarder = group

        if self.actuator is None:
            raise ValueError('Lack of Action object')
        if self.encoder is None:
            raise ValueError('Lack of Encoder object')

        self.network.build(self._backend)

        self.state = np.zeros(self.encoder.num)
        self.environment = environment
        if self.environment.shape is not None and len(self.environment.shape) >= 2:
            self.conv_state = True
        else:
            self.conv_state = False
        self.step_count = 0
        self.episode = 0
        self.num_episodes = kwargs.get('num_episodes', 100)

        self.accumulated_reward = 0.0
        self.reward_list = []
        self.action = -1
        self.last_action = -1

        self.action_repeat_count = 0
        self.action_repeat = kwargs.get('action_repeat', 2)

        self.probability_random_action = kwargs.get('probability_random_action', 0.0)
        self.render_interval = kwargs.get('render_interval', None)
        self.reward_delay = kwargs.get('reward_delay', None)
        self.replay_memory = kwargs.get('replay_memory', False)
        if self.replay_memory:
            self.memory_capacity = kwargs.get('memory_capacity', 10000)
            self.memory_pool = ReplayMemory(self.memory_capacity)

        if self.reward_delay is not None:
            assert self.reward_delay > 0
            self.rewards = np.zeros(self.reward_delay)

    def env_step(self):
        """
        Single step of the environment which includes rendering, getting and performing
        the action, and accumulating/delaying rewards.

        Returns:
            An OpenAI ``gym`` compatible tuple (next_state, reward, done).
        """
        # Render the environment.
        if (self.render_interval is not None and self.step_count % self.render_interval == 0):
            self.environment.render()

        # Get action
        self.last_action = self.action
        # action sampled from the action space with a certain probability
        if np.random.rand(1) < self.probability_random_action:
            self.action = np.random.randint(
                low=0, high=self.environment.action_num, size=(1,)
            )[0]
        elif self.action_repeat_count > self.action_repeat:
            if self.last_action == 0:
                self.action = 1
                tqdm.write(f"Act -> too many times {self.last_action} ")
            else:
                self.action = np.random.randint(
                    low=0, high=self.environment.action_num, size=(1,)
                )[0]
                tqdm.write(f"Act -> too many times {self.last_action} ")
        else:
            if self.sim_name == 'pytorch':
                self.action = int(self.actuator.action)  # Get action from the predict result of action

        if self.last_action == self.action:
            self.action_repeat_count += 1
        else:
            self.action_repeat_count = 0

        # Run a step of the environment.
        next_state, reward, done, _ = self.environment.step(self.action)
        if self.conv_state:
            next_state = next_state[np.newaxis, :]

        if done:
            next_state = None

        # Set reward in case of delay.
        if self.reward_delay is not None:
            if self.sim_name == 'pytorch':
                self.rewards = torch.tensor([reward, *self.rewards[1:]], device=self.device)
            reward = self.rewards[-1]

        # Accumulate reward.
        self.accumulated_reward += reward
        if self.replay_memory:
            self.memory_pool.push(self.state, self.action, next_state, reward)
        self.state = next_state

        return next_state, reward, done

    def update_step(self, gym_batch, **kwargs):
        """
        Run a single iteration of the network and update it and the reward list when
        done.

        Args:
            gym_batch (tuple): An OpenAI ``gym`` compatible tuple (next_state, reward, done).
        """
        self.step_count += 1
        next_state, reward, done = gym_batch

        if done:
            self.reward_list.append(self.accumulated_reward)
            print("Episode finished after {} steps".format(self.step_count))
        else:
            # Add a placeholder for batch_size
            next_state = next_state[np.newaxis, :]

            # Place the observations into the network.
            # self.network.input(next_state)
            self.encoder(next_state)
            if self.rewarder is not None:
                self.rewarder(reward)

            self.network.run(self.time)

    def reset_pipeline(self):
        """
        Reset the pipeline.
        """
        self.environment.reset()
        self.accumulated_reward = 0.0
        self.step_count = 0
        self.action = -1
        self.last_action = -1
        self.action_repeat_count = 0
        if self.conv_state:
            self.state = np.zeros(self.environment.shape)
            self.state = self.state[np.newaxis, :]
        else:
            self.state = np.zeros(self.encoder.num)


Transition = namedtuple('Transition', ('state', 'action', 'next_state', 'reward'))


class ReplayMemory(object):

    def __init__(self, capacity):
        self.memory = deque([], maxlen=capacity)

    def push(self, *args):
        """Save a transition"""
        self.memory.append(Transition(*args))

    def sample(self, batch_size):
        return random.sample(self.memory, batch_size)

    def __len__(self):
        return len(self.memory)







# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Initializer.py
@time:2022/8/11 13:59
@description:
"""
__all__ = ['uniform', 'normal', 'xavier_normal', 'xavier_uniform', 'kaiming_normal', 'kaiming_uniform', 'constant', 'sparse']


class BaseInitializer(object):
    r"""
    """
    def __init__(self, **kwargs):
        super().__init__()


class uniform(BaseInitializer):
    def __init__(self, a=-0.0, b=1.0):
        super(uniform, self).__init__(a=a, b=b)
        '''
        Args:
            a(float): the lower bound of the uniform distribution
            b(float): the upper bound of the uniform distribution
        '''
        self.a = a
        self.b = b


class normal(BaseInitializer):
    def __init__(self, mean=0.0, std=1.0):
        super(normal, self).__init__(mean=mean, std=std)
        '''
        Args:
            mean(float): the mean of the normal distribution
            std(float): the standard deviation of the normal distribution
        Returns:
        '''
        self.mean = mean
        self.std = std


class xavier_normal(BaseInitializer):
    def __init__(self, gain=1.0):
        super(xavier_normal, self).__init__(gain=gain)
        '''
        Args:
            gain: an optional scaling factor
        '''
        self.gain = gain
        # return torch.nn.init.xavier_normal_(data, gain)


class xavier_uniform(BaseInitializer):
    def __init__(self, gain=1.0):
        super(xavier_uniform, self).__init__(gain=gain)
        '''
        Args:
            gain: an optional scaling factor
        '''
        self.gain = gain
        # return torch.nn.init.xavier_uniform_(data, gain)


class kaiming_normal(BaseInitializer):
    def __init__(self, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        super(kaiming_normal, self).__init__(a=a, mode=mode, nonlinearity=nonlinearity)
        '''
        Args:
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of 
            the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 
            'leaky_relu' (default).
       
        '''
        self.a = a
        self.mode = mode
        self.nonlinearity = nonlinearity
        # return torch.nn.init.kaiming_normal_(data, a, mode, nonlinearity)


class kaiming_uniform(BaseInitializer):
    def __init__(self, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        super(kaiming_uniform, self).__init__(a=a, mode=mode, nonlinearity=nonlinearity)
        '''
        Args:
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance 
            of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 
            'leaky_relu' (default).
        '''
        self.a = a
        self.mode = mode
        self.nonlinearity = nonlinearity
        # return torch.nn.init.kaiming_uniform_(data, a, mode, nonlinearity)


class constant(BaseInitializer):
    def __init__(self, constant_value=0.0):
        super(constant, self).__init__(constant_value=constant_value)
        '''
        Args:
            constant_value(float): the value to fill the tensor with
        '''
        self.constant_value = constant_value
        # return torch.nn.init.constant_(data, constant_value)


class sparse(BaseInitializer):
    def __init__(self, sparsity=0.1, std=0.01):
        super(sparse, self).__init__(sparsity=sparsity, std=std)
        '''
        Args:
            sparsity(float): The fraction of elements in each column to be set to zero
            std(float): the standard deviation of the normal distribution used to generate
            the non-zero values
        Returns:
            torch.nn.init.sparse_(data, sparsity, std)
        '''
        self.sparsity = sparsity
        self.std = std
        # return torch.nn.init.sparse_(data, sparsity, std)
